from egcg_core.app_logging import AppLogger


class Notification(AppLogger):
    def __init__(self, name):
        self.name = name

    def notify(self, msg, attachments=None):
        raise NotImplementedError



import json
import os
import shutil

import requests
from os import getenv
from time import sleep
from datetime import datetime
from unittest import TestCase
from multiprocessing import Lock
from subprocess import check_output
from egcg_core import config, rest_communication


def get_cfg():
    cfg = config.Configuration()
    cfg_file = getenv('INTEGRATIONCONFIG')
    if cfg_file:
        cfg.load_config_file(cfg_file)

    return cfg


def now():
    return datetime.utcnow().strftime('%Y-%m-%d_%H:%M:%S')


class WrappedFunc:
    """For wrapping unittest.TestCase's assert methods, logging details of the assertion carried out."""
    def __init__(self, test_case, assert_func):
        self.test_case = test_case
        self.assert_func = assert_func
        self.lock = Lock()

    def __call__(self, check_name, *args, **kwargs):
        if not isinstance(check_name, str):
            raise NameError('Incorrect call - check_name required')

        assertion_report = '%s.%s\t%s\t%s\t' % (
            self.test_case.__class__.__name__,
            self.test_case._testMethodName,
            check_name,
            self.assert_func.__name__
        )
        args_used = '\t' + str(args)
        if kwargs:
            args_used += ' ' + str(kwargs)

        try:
            self.assert_func(*args, **kwargs)
            self.log(assertion_report + 'success' + args_used)
        except AssertionError:
            self.log(assertion_report + 'failed' + args_used)
            raise

    def log(self, msg):
        with self.lock:
            with open('checks.log', 'a') as f:
                f.write(msg + '\n')


class IntegrationTest(TestCase):
    """Contains some common functionality for patching, and quality-oriented assertion logging."""
    _wrapped_func_blacklist = ('assertRaises', 'assertWarns', 'assertLogs', 'assertRaisesRegex', 'assertWarnsRegex')
    patches = ()

    def __init__(self, *args):
        super().__init__(*args)

        # need separate TestCase with non-wrapped assert methods, because some of them call each other
        self.asserter = TestCase()
        for attrname in dir(self.asserter):
            if attrname.startswith('assert'):
                attr = getattr(self.asserter, attrname)
                if callable(attr) and attrname not in self._wrapped_func_blacklist:
                    setattr(self, attrname, WrappedFunc(self, attr))

        self.cfg = get_cfg()

    def setUp(self):
        for p in self.patches:
            p.start()

    def tearDown(self):
        for p in self.patches:
            p.stop()


class ReportingAppIntegrationTest(IntegrationTest):
    """
    Sets up a Reporting-App Docker image before each test, and stops/removes it after. Captures the image's IP address
    and uses that to patch egcg_core.rest_communication.default.
    """
    container_id = None
    container_ip = None
    container_port = None
    lims_data_yaml = None
    users_sqlite = None
    mongo_db = None

    @property
    def reporting_app_data(self):
        """
        directory containing the reporting app data that will be mounted on the docker image.
        Location depends on the current working directory.
        """
        return os.path.abspath('reporting_app_data')

    def _loadup_directory_with_data(self):
        """Prepare the directory that will be passed on to reporting app docker image."""
        os.makedirs(self.reporting_app_data, exist_ok=True)
        lims_data_yaml = self.cfg.query('reporting_app', 'lims_data_yaml', ret_default=self.lims_data_yaml)
        if lims_data_yaml and os.path.isfile(lims_data_yaml):
            shutil.copyfile(lims_data_yaml, os.path.join(self.reporting_app_data, 'data_for_clarity_lims.yaml'))
        users_sqlite = self.cfg.query('reporting_app', 'users_sqlite', ret_default=self.users_sqlite)
        if users_sqlite and os.path.isfile(users_sqlite):
            shutil.copyfile(users_sqlite, os.path.join(self.reporting_app_data, 'users.sqlite'))
        mongo_db = self.cfg.query('reporting_app', 'mongo_db', ret_default=self.mongo_db)
        if mongo_db and os.path.isdir(mongo_db):
            shutil.copytree(mongo_db, os.path.join(self.reporting_app_data, 'db'))

    def setUp(self):
        super().setUp()
        self._loadup_directory_with_data()

        self.container_id = check_output(
            ['docker', 'run', '-d', '-v', self.reporting_app_data + ':/opt/etc',
             self.cfg['reporting_app']['image_name'],
             self.cfg.query('reporting_app', 'branch', ret_default='master')]
        ).decode().strip()
        assert self.container_id

        container_basename = self.cfg.query('reporting_app', 'container_basename')
        if container_basename:
            check_output(['docker', 'rename', self.container_id, container_basename + '_' + self.container_id[:12]])

        container_info = json.loads(check_output(['docker', 'inspect', self.container_id]).decode())[0]
        # for now, assume the container is running on the main 'bridge' network
        self.container_ip = container_info['NetworkSettings']['Networks']['bridge']['IPAddress']
        self.container_port = list(container_info['Config']['ExposedPorts'])[0].rstrip('/tcp')
        container_url = 'http://' + self.container_ip + ':' + self.container_port + '/api/0.1'
        rest_communication.default._baseurl = container_url
        rest_communication.default._auth = (self.cfg['reporting_app']['username'], self.cfg['reporting_app']['password'])
        self._ping(container_url)

    def tearDown(self):
        super().tearDown()

        assert self.container_id
        check_output(['docker', 'stop', self.container_id])
        check_output(['docker', 'rm', '-v', self.container_id])
        shutil.rmtree(self.reporting_app_data)
        self.container_id = self.container_ip = self.container_port = None

    def _ping(self, url, retries=36):
        try:
            requests.get(url, timeout=2)
            return True
        except requests.exceptions.ConnectionError:
            if retries > 0:
                sleep(5)
                return self._ping(url, retries - 1)
            else:
                raise



import json
import os.path
import unittest
from unittest.mock import Mock
from egcg_core.config import cfg


class FakeRestResponse(Mock):
    request = Mock(method='a method', path_url='a url')
    status_code = 200
    reason = 'a reason'

    def __init__(self, content, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.content = json.dumps(content).encode()

    def json(self):
        return json.loads(self.text)

    @property
    def text(self):
        return self.content.decode('utf-8')


class TestEGCG(unittest.TestCase):
    file_path = os.path.dirname(__file__)
    assets_path = os.path.join(file_path, 'assets')
    etc = os.path.join(os.path.dirname(file_path), 'etc')
    etc_config = os.path.join(etc, 'example_egcg.yaml')

    @classmethod
    def setUpClass(cls):
        cfg.load_config_file(cls.etc_config)



import sqlite3
from unittest.mock import patch
from tests import FakeRestResponse
from egcg_core import ncbi


def reset_cache():
    ncbi.data_cache = sqlite3.connect(':memory:')
    ncbi.cursor = ncbi.data_cache.cursor()
    ncbi._create_tables()


def test_fetch_from_eutils():
    ncbi_search_data = {'esearchresult': {'idlist': ['1337']}}
    ncbi_fetch_data = '''
        <ScientificName>Genus species</ScientificName>
        <OtherNames><CommonName>a common name</CommonName></OtherNames>
        <Rank>species</Rank>
    '''
    ncbi_fetch_data_sub_species = '''
        <ScientificName>Genus species</ScientificName>
        <OtherNames><CommonName>a common name</CommonName></OtherNames>
        <Rank>subspecies</Rank>
    '''

    patched_get = patch(
        'egcg_core.ncbi.requests.get',
        side_effect=(
            FakeRestResponse(ncbi_search_data),
            FakeRestResponse(ncbi_fetch_data),
            FakeRestResponse(ncbi_fetch_data),
            FakeRestResponse(ncbi_fetch_data)
        )
    )
    patched_get2 = patch(
        'egcg_core.ncbi.requests.get',
        side_effect=(
            FakeRestResponse(ncbi_search_data),
            FakeRestResponse(ncbi_fetch_data_sub_species),
            FakeRestResponse(ncbi_fetch_data_sub_species),
            FakeRestResponse(ncbi_fetch_data_sub_species)
        )
    )
    with patched_get as mocked_get:
        obs = ncbi.fetch_from_eutils('a_species')
        assert obs == ('1337', 'Genus species', 'a common name')
        mocked_get.assert_any_call(
            'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi',
            params={'db': 'Taxonomy', 'term': 'a_species', 'retmode': 'JSON'}
        )
        mocked_get.assert_any_call(
            'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi',
            params={'db': 'Taxonomy', 'id': '1337'}
        )
    with patched_get2:
        obs = ncbi.fetch_from_eutils('a_species')
        assert obs == ('1337', 'Genus species', 'a common name')


def test_cache():
    assert ncbi.fetch_from_cache('a species') is None
    ncbi._cache_species('a species', '1337', 'Scientific name', 'a species')
    assert ncbi.fetch_from_cache('a species') == ('a species', '1337', 'Scientific name', 'a species')
    reset_cache()


@patch('egcg_core.ncbi.fetch_from_eutils')
def test_get_species_name(mocked_fetch):
    for table in ('species', 'aliases'):
        ncbi.cursor.execute('SELECT * FROM ' + table)
        assert ncbi.cursor.fetchall() == []

    mocked_fetch.return_value = (None, None, None)
    assert ncbi.get_species_name('a species') is None
    for table in ('species', 'aliases'):
        ncbi.cursor.execute('SELECT * FROM ' + table)
        assert ncbi.cursor.fetchall() == []

    reset_cache()
    mocked_fetch.return_value = ('1337', 'Scientific name', 'a species')
    assert ncbi.get_species_name('a species') == 'Scientific name'
    assert ncbi.fetch_from_cache('a species') == ('a species', '1337', 'Scientific name', 'a species')
    reset_cache()



import asana
from cached_property import cached_property
from os.path import basename
from .notification import Notification


class AsanaNotification(Notification):
    def __init__(self, name, workspace_id, project_id, access_token, task_description=None):
        super().__init__(name)
        self.task_id = self.name
        self.workspace_id = workspace_id
        self.project_id = project_id
        self.client = asana.Client.access_token(access_token)
        self.task_template = {'name': name, 'projects': [self.project_id]}
        if task_description:
            self.task_template['notes'] = task_description

    def notify(self, msg, attachments=None):
        self.client.tasks.add_comment(self.task['gid'], text=msg)
        self.client.tasks.update(self.task['gid'], completed=False)

        if attachments:
            if isinstance(attachments, str):
                attachments = [attachments]

        for attachment in attachments or []:
            with open(attachment, 'rb') as fil:
                content = fil.read()

                # can't use self.client.attachments.create_on_task here - results in TypeError
                self.client.request(
                    'post',
                    '/tasks/%s/attachments' % self.task['gid'],
                    files=[('file', (basename(attachment), content, None))],
                )

    @cached_property
    def task(self):
        tasks = list(self.client.tasks.find_all(project=self.project_id))
        task_ent = self._get_entity(tasks, self.task_id)
        if task_ent is None:
            task_ent = self._create_task()
        return self.client.tasks.find_by_id(task_ent['gid'])

    @staticmethod
    def _get_entity(collection, name):
        for e in collection:
            if e['name'] == name:
                return e

    def _create_task(self):
        return self.client.tasks.create_in_workspace(self.workspace_id, self.task_template)



import multiprocessing
import os
import json

import pytest
from requests import Session
from requests.exceptions import SSLError
from unittest.mock import MagicMock, patch, call

from tests import FakeRestResponse, TestEGCG
from egcg_core import rest_communication
from egcg_core.util import check_if_nested
from egcg_core.exceptions import RestCommunicationError


test_endpoint = 'an_endpoint'
test_url = 'http://localhost:4999/api/0.1/' + test_endpoint + '/'
test_nested_request_content = {'data': ['some', {'test': 'content'}]}
test_flat_request_content = {'key1': 'value1', 'key2': 'value2'}
test_patch_document = {
    '_id': '1337', '_etag': 1234567, 'uid': 'a_unique_id', 'list_to_update': ['this', 'that', 'other']
}


def fake_request(method, url, **kwargs):
    if kwargs.get('files'):
        if 'json' in kwargs:
            raise Exception
        if 'data' in kwargs and check_if_nested(kwargs['data']):
            raise Exception
    return FakeRestResponse(test_nested_request_content)


patched_request = patch.object(Session, 'request', side_effect=fake_request)
patched_failed_request = patch.object(Session, 'request', side_effect=SSLError('SSL error'))
auth = ('a_user', 'a_password')


class TestRestCommunication(TestEGCG):
    def setUp(self):
        self.comm = rest_communication.Communicator(auth=auth, baseurl='http://localhost:4999/api/0.1')

    def test_begin_session(self):
        s = self.comm.begin_session()
        assert s.adapters['http://'] is s.adapters['https://']
        assert s.auth == auth
        assert s.params == {}

        hashed_token = '{"some": "hashed"}.tokenauthentication'
        self.comm._auth = hashed_token
        s = self.comm.begin_session()
        assert s.headers['Authorization'] == 'Token ' + hashed_token

    def test_api_url(self):
        assert self.comm.api_url('an_endpoint') == test_url

    def test_parse_query_string(self):
        query_string = 'http://a_url?this=that&other={"another":"more"}&things=1'
        dodgy_query_string = 'http://a_url?this=that?other=another'

        assert self.comm._parse_query_string('http://a_url') == {}
        assert self.comm._parse_query_string(query_string) == {
            'this': 'that', 'other': '{"another":"more"}', 'things': '1'
        }

        with self.assertRaises(RestCommunicationError) as e:
            self.comm._parse_query_string(dodgy_query_string)
            assert str(e.exception) == 'Bad query string: ' + dodgy_query_string

        with self.assertRaises(RestCommunicationError) as e2:
            self.comm._parse_query_string(query_string, requires=['thangs'])
            assert str(e2.exception) == query_string + " did not contain all required fields: ['thangs']"

    def test_detect_files_in_json(self):
        json_no_files = {'k1': 'v1', 'k2': 'v2'}
        obs_files, obs_json = self.comm._detect_files_in_json(json_no_files)
        assert obs_files is None
        assert obs_json == json_no_files

        file_path = os.path.join(self.assets_path, 'test_to_upload.txt')
        json_with_files = {'k1': 'v1', 'k2': ('file', file_path)}
        obs_files, obs_json = self.comm._detect_files_in_json(json_with_files)
        assert obs_files == {'k2': (file_path, b'test content', 'text/plain')}
        assert obs_json == {'k1': 'v1'}

        json_list = [json_with_files, json_with_files]
        obs_files, obs_json = self.comm._detect_files_in_json(json_list)
        assert obs_files == [
            {'k2': (file_path, b'test content', 'text/plain')},
            {'k2': (file_path, b'test content', 'text/plain')}
        ]
        assert obs_json == [{'k1': 'v1'}, {'k1': 'v1'}]

    @patched_request
    def test_req(self, mocked_request):
        json_content = ['some', {'test': 'json'}]
        response = self.comm._req('METHOD', test_url, json=json_content)
        assert response.status_code == 200
        assert json.loads(response.content.decode('utf-8')) == response.json() == test_nested_request_content
        mocked_request.assert_called_with('METHOD', test_url, json=json_content)

    @patch('egcg_core.rest_communication.sleep')
    @patch.object(Session, 'request', side_effect=[SSLError('Error 1'), SSLError('Error 2'), FakeRestResponse(test_nested_request_content)])
    def test_retry(self, mocked_request, mocked_sleep):
        assert self.comm._req('METHOD', test_url, json=[]).json() == test_nested_request_content
        c = call('METHOD', test_url, json=[])
        mocked_request.assert_has_calls([c, c, c])
        assert mocked_sleep.call_count == 2

    @patch('egcg_core.rest_communication.sleep')
    @patched_failed_request
    def test_failed_req(self, mocked_request, mocked_sleep):
        json_content = ['some', {'test': 'json'}]
        self.comm.lock = MagicMock()
        self.comm.lock.__enter__.assert_not_called()
        self.comm.lock.__exit__.assert_not_called()

        with pytest.raises(SSLError):
            _ = self.comm._req('METHOD', test_url, json=json_content)

        assert self.comm.lock.__enter__.call_count == 6
        assert self.comm.lock.__exit__.call_count == 6  # exception raised, but lock still released
        mocked_sleep.assert_called_with(2)
        assert mocked_sleep.call_count == 5

    @patched_request
    def test_multi_session(self, mocked_request):
        json_content = ['some', {'test': 'json'}]
        with patch('os.getpid', return_value=1):
            _ = self.comm._req('METHOD', test_url, json=json_content)
        with patch('os.getpid', return_value=2):
            _ = self.comm._req('METHOD', test_url, json=json_content)
        assert len(self.comm._sessions) == 2

    @patched_request
    def test_with_multiprocessing(self, mocked_request):
        json_content = ['some', {'test': 'json'}]

        def assert_request():
            _ = self.comm._req('METHOD', test_url, json=json_content)
            assert mocked_request.call_count == 2
            assert len(self.comm._sessions) == 2

        # initiate in the Session in the main thread
        self.comm._req('METHOD', test_url, json=json_content)
        procs = []
        for i in range(10):
            procs.append(multiprocessing.Process(target=assert_request))
        for p in procs:
            p.start()
        for p in procs:
            p.join()

    @patch.object(Session, '__exit__')
    @patch.object(Session, '__enter__')
    @patched_request
    def test_context_manager(self, mocked_request, mocked_enter, mocked_exit):
        json_content = ['some', {'test': 'json'}]
        with self.comm.session:
            mocked_enter.assert_called_once()
            mocked_exit.assert_not_called()
            for i in range(4):  # multiple calls
                response = self.comm._req('METHOD', test_url, json=json_content)
                assert response.status_code == 200
                assert response.json() == test_nested_request_content
                mocked_request.assert_called_with('METHOD', test_url, json=json_content)

        assert mocked_request.call_count == 4
        mocked_exit.assert_called_once()

    @patch.object(rest_communication.Communicator, 'error')
    @patch.object(Session, 'request')
    def test_communication_error(self, mocked_req, mocked_log):
        response = FakeRestResponse({})
        response.status_code = 500
        mocked_req.return_value = response
        self.comm.lock = MagicMock()
        self.comm.lock.__enter__.assert_not_called()
        self.comm.lock.__exit__.assert_not_called()

        with self.assertRaises(RestCommunicationError) as e:
            self.comm.get_document('an_endpoint')

        assert mocked_log.call_args[0][0].endswith('Status code 500. Reason: a reason')
        assert str(e.exception) == 'Encountered a 500 status code: a reason'
        self.comm.lock.__enter__.assert_called_once()
        self.comm.lock.__exit__.assert_called_once()  # exception raised, but lock still released

    @patch.object(rest_communication.Communicator, '_req')
    def test_get_documents_depaginate(self, mocked_req):
        mocked_req.side_effect = (
            FakeRestResponse({'data': ['this', 'that'], '_links': {'next': {'href': 'an_endpoint?max_results=101&page=2'}}}),
            FakeRestResponse({'data': ['other', 'another'], '_links': {'next': {'href': 'an_endpoint?max_results=101&page=3'}}}),
            FakeRestResponse({'data': ['more', 'things'], '_links': {}})
        )

        assert self.comm.get_documents('an_endpoint', all_pages=True, max_results=101) == [
            'this', 'that', 'other', 'another', 'more', 'things'
        ]
        mocked_req.assert_has_calls(
            (
                # Communicator.get_content passes ints
                call('GET', test_url, params={'page': 1, 'max_results': 101}, quiet=False),
                # url parsing passes strings, but requests removes the quotes anyway
                call('GET', test_url, params={'page': '2', 'max_results': '101'}, quiet=False),
                call('GET', test_url, params={'page': '3', 'max_results': '101'}, quiet=False)
            )
        )

        docs = [
            FakeRestResponse(
                {
                    'data': ['data%s' % d],
                    '_links': {'next': {'href': 'an_endpoint?max_results=101&page=%s' % d}}
                }
            )
            for d in range(1, 1200)
        ]
        docs.append(FakeRestResponse({'data': ['last piece'], '_links': {}}))

        mocked_req.side_effect = docs
        ret = self.comm.get_documents('an_endpoint', all_pages=True, max_results=101)
        assert len(ret) == 1200

    @patched_request
    def test_get_content(self, mocked_request):
        data = self.comm.get_content(test_endpoint, max_results=100, where={'a_field': 'thing'})
        assert data == test_nested_request_content
        mocked_request.assert_called_with(
            'GET',
            test_url,
            params={'max_results': 100, 'where': '{"a_field": "thing"}', 'page': 1}
        )

    def test_get_documents(self):
        with patched_request:
            data = self.comm.get_documents(test_endpoint, max_results=100, where={'a_field': 'thing'})
            assert data == test_nested_request_content['data']

    def test_get_document(self):
        expected = test_nested_request_content['data'][0]
        with patched_request:
            observed = self.comm.get_document(test_endpoint, max_results=100, where={'a_field': 'thing'})
            assert observed == expected

    @patched_request
    def test_post_entry(self, mocked_request):
        self.comm.post_entry(test_endpoint, payload=test_nested_request_content)
        mocked_request.assert_called_with(
            'POST',
            test_url,
            json=test_nested_request_content,
            files=None
        )
        file_path = os.path.join(self.assets_path, 'test_to_upload.txt')
        test_request_content_plus_files = dict(test_flat_request_content)
        test_request_content_plus_files['f'] = ('file', file_path)
        self.comm.post_entry(test_endpoint, payload=test_request_content_plus_files)
        mocked_request.assert_called_with(
            'POST',
            test_url,
            data=test_flat_request_content,
            files={'f': (file_path, b'test content', 'text/plain')}
        )

        self.comm.post_entry(test_endpoint, payload=test_flat_request_content, use_data=True)
        mocked_request.assert_called_with(
            'POST',
            test_url,
            data=test_flat_request_content,
            files=None
        )

        self.comm.post_entry(test_endpoint, payload=test_request_content_plus_files, use_data=True)
        mocked_request.assert_called_with(
            'POST',
            test_url,
            data=test_flat_request_content,
            files={'f': (file_path, b'test content', 'text/plain')}
        )

    @patched_request
    def test_put_entry(self, mocked_request):
        self.comm.put_entry(test_endpoint, 'an_element_id', payload=test_nested_request_content)
        mocked_request.assert_called_with(
            'PUT',
            test_url + 'an_element_id',
            json=test_nested_request_content,
            files=None
        )

        file_path = os.path.join(self.assets_path, 'test_to_upload.txt')
        test_request_content_plus_files = dict(test_flat_request_content)
        test_request_content_plus_files['f'] = ('file', file_path)
        self.comm.put_entry(test_endpoint, 'an_element_id', payload=test_request_content_plus_files)
        mocked_request.assert_called_with(
            'PUT',
            test_url + 'an_element_id',
            data=test_flat_request_content,
            files={'f': (file_path, b'test content', 'text/plain')}
        )

    @patch.object(rest_communication.Communicator, 'get_document', return_value=test_patch_document)
    @patched_request
    def test_patch_entry(self, mocked_request, mocked_get_doc):
        patching_payload = {'list_to_update': ['another']}
        self.comm.patch_entry(
            test_endpoint,
            payload=patching_payload,
            id_field='uid',
            element_id='a_unique_id',
            update_lists=['list_to_update']
        )

        mocked_get_doc.assert_called_with(test_endpoint, where={'uid': 'a_unique_id'})
        mocked_request.assert_called_with(
            'PATCH',
            test_url + '1337',
            headers={'If-Match': 1234567},
            json={'list_to_update': ['this', 'that', 'other', 'another']},
            files=None
        )

    @patch.object(rest_communication.Communicator, 'get_document', return_value=test_patch_document)
    @patched_request
    def test_if_match(self, mocked_request, mocked_get_doc):
        self.comm.patch_entry(test_endpoint, {'this': 'that'}, 'uid', 'a_unique_id')
        mocked_get_doc.assert_called_with(test_endpoint, where={'uid': 'a_unique_id'})
        mocked_request.assert_called_with(
            'PATCH',
            test_url + '1337',
            headers={'If-Match': 1234567},
            json={'this': 'that'},
            files=None
        )

    @patch.object(rest_communication.Communicator, 'post_entry', return_value=True)
    @patch.object(rest_communication.Communicator, '_patch_entry', return_value=True)
    @patch.object(rest_communication.Communicator, 'get_document')
    def test_post_or_patch(self, mocked_get, mocked_patch, mocked_post):
        test_post_or_patch_payload = {'uid': '1337', 'list_to_update': ['more'], 'another_field': 'that'}
        test_post_or_patch_payload_no_uid = {'list_to_update': ['more'], 'another_field': 'that'}
        test_post_or_patch_doc = {
            'uid': 'a_uid', '_id': '1337', '_etag': 1234567, 'list_to_update': ['things'], 'another_field': 'this'
        }
        mocked_get.return_value = test_post_or_patch_doc

        self.comm.post_or_patch(
            'an_endpoint',
            [test_post_or_patch_payload],
            id_field='uid',
            update_lists=['list_to_update']
        )
        mocked_get.assert_called_with('an_endpoint', where={'uid': '1337'})
        mocked_patch.assert_called_with(
            'an_endpoint',
            test_post_or_patch_doc,
            test_post_or_patch_payload_no_uid,
            ['list_to_update']
        )

        mocked_get.return_value = None
        self.comm.post_or_patch(
            'an_endpoint', [test_post_or_patch_payload], id_field='uid', update_lists=['list_to_update']
        )
        mocked_get.assert_called_with('an_endpoint', where={'uid': '1337'})
        mocked_post.assert_called_with('an_endpoint', test_post_or_patch_payload)


def test_default():
    d = rest_communication.default
    assert d.baseurl == 'http://localhost:4999/api/0.1'
    assert d.auth == ('a_user', 'a_password')



import sys
import traceback
from egcg_core.app_logging import AppLogger
from egcg_core.config import cfg
from egcg_core.exceptions import EGCGError
from .asana import AsanaNotification
from .email import EmailNotification, EmailSender, send_email, send_html_email, send_plain_text_email
from .log import LogNotification


class NotificationCentre(AppLogger):
    ntf_aliases = {
        'log': LogNotification,
        'email': EmailNotification,
        'asana': AsanaNotification
    }

    def __init__(self, name):
        self.name = name
        self.subscribers = {}

        for k, v in cfg.get('notifications', {}).items():
            if k in self.ntf_aliases:
                self.debug('Configuring notification for: ' + k)
                self.subscribers[k] = self.ntf_aliases[k](name=self.name, **v)
            else:
                self.warning("Bad notification config '%s' - this will be ignored", k)

    def notify(self, msg, subs):
        exceptions = []
        for s in subs:
            if s in self.subscribers:
                try:
                    self.subscribers[s].notify(msg)
                except Exception as e:
                    etype, value, tb = sys.exc_info()
                    stacktrace = ''.join(traceback.format_exception(etype, value, tb))
                    self.critical(stacktrace)
                    exceptions.append(e)
            else:
                self.debug('Tried to notify by %s, but no configuration present', s)

        if exceptions:
            raise EGCGError(
                'Encountered the following errors during notification: %s' % ', '.join(
                    '%s: %s' % (e.__class__.__name__, str(e))
                    for e in exceptions
                )
            )

    def notify_all(self, msg):
        self.notify(msg, self.subscribers.keys())



import re
import sqlite3
import requests
from egcg_core.config import cfg
from egcg_core.app_logging import logging_default as log_cfg

app_logger = log_cfg.get_logger(__name__)

data_cache = None
cursor = None


def _connect():
    global data_cache
    global cursor

    data_cache = sqlite3.connect(cfg['ncbi_cache'])
    cursor = data_cache.cursor()


def _create_tables():
    _create = 'CREATE TABLE IF NOT EXISTS '
    cursor.execute(_create + 'species (taxid text UNIQUE, scientific_name text UNIQUE, common_name text)')
    cursor.execute(_create + 'aliases (query_name text UNIQUE, taxid text REFERENCES species(taxid))')


def connect():
    if data_cache is None and cursor is None:
        _connect()
        _create_tables()


def get_species_name(query_species):
    local_query = fetch_from_cache(query_species)
    if local_query:
        q, taxid, scientific_name, common_name = local_query
    else:
        taxid, scientific_name, common_name = fetch_from_eutils(query_species)
        if taxid is None:
            return None

        _cache_species(query_species, taxid, scientific_name, common_name)

    return scientific_name


def fetch_from_cache(query_species):
    connect()
    cursor.execute('SELECT * FROM aliases NATURAL JOIN species WHERE query_name=?', (query_species,))
    return cursor.fetchone()


def _cache_species(query_species, taxid, scientific_name, common_name):
    connect()
    cursor.execute('SELECT taxid FROM species WHERE taxid=?', (taxid,))
    if not cursor.fetchone():
        cursor.execute('INSERT INTO species VALUES (?, ?, ?)', (taxid, scientific_name, common_name))
    cursor.execute('INSERT INTO aliases VALUES (?, ?)', (query_species, taxid))
    data_cache.commit()


def fetch_from_eutils(species):
    """
    Query NCBI taxomomy database to get the taxonomy ID, scientific name and common name.
    Documentation available at http://www.ncbi.nlm.nih.gov/books/NBK25499/
    :param str species: A search term for esearch, e.g. Human, Mouse, etc.
    """
    esearch_url = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi'
    payload = {'db': 'Taxonomy', 'term': species, 'retmode': 'JSON'}
    r = requests.get(esearch_url, params=payload)
    results = r.json()
    taxid_list = results.get('esearchresult').get('idlist')
    all_species_names = []
    for taxid in taxid_list:
        efetch_url = 'http://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi'
        payload = {'db': 'Taxonomy', 'id': taxid}
        r = requests.get(efetch_url, params=payload)
        match = re.search('<Rank>(.+?)</Rank>', r.text, re.MULTILINE)
        rank = None
        if match:
            rank = match.group(1)
        if rank in ['species', 'subspecies']:
            scientific_name = common_name = None
            match = re.search('<ScientificName>(.+?)</ScientificName>', r.text, re.MULTILINE)
            if match:
                scientific_name = match.group(1)
            match = re.search('<GenbankCommonName>(.+?)</GenbankCommonName>', r.text, re.MULTILINE)
            if not match:
                match = re.search('<CommonName>(.+?)</CommonName>', r.text, re.MULTILINE)
            if match:
                common_name = match.group(1)
            all_species_names.append((taxid, scientific_name, common_name))
    nspecies = len(all_species_names)
    if nspecies != 1:
        app_logger.error('%s taxons found for %s', nspecies, species)
        return None, None, None

    return all_species_names[0]


# backward compatibility
_fetch_from_cache = fetch_from_cache
_fetch_from_eutils = fetch_from_eutils



import filecmp
import os
import shutil

import requests
from unittest.mock import Mock, patch
from egcg_core import integration_testing, config
from tests import TestEGCG


class TestIntegrationTesting(TestEGCG):
    def setUp(self):
        self.original_baseurl = integration_testing.rest_communication.default.baseurl

    def tearDown(self):
        integration_testing.rest_communication.default._baseurl = self.original_baseurl

    def test_wrapped_func(self):
        if os.path.isfile('checks.log'):
            os.remove('checks.log')
        w = integration_testing.WrappedFunc(self, self.assertEqual)
        w('a check', 1, 1)
        with self.assertRaises(AssertionError):
            w('another check', 1, 2)

        with open('checks.log', 'r') as f:
            assert f.readline() == '%s.test_wrapped_func\ta check\tassertEqual\tsuccess\t(1, 1)\n' % self.__class__.__name__
            assert f.readline() == '%s.test_wrapped_func\tanother check\tassertEqual\tfailed\t(1, 2)\n' % self.__class__.__name__

    def test_integration_test(self):
        t = integration_testing.IntegrationTest()
        t.patches = (patch('os.path.isdir', return_value=1337),)
        for attr in ('assertEqual', 'assertTrue'):
            assert getattr(t, attr).assert_func == getattr(t.asserter, attr)

        wd = os.getcwd()
        assert os.path.isdir(wd) is True
        t.setUp()
        assert os.path.isdir(wd) == 1337
        t.tearDown()
        assert os.path.isdir(wd) is True

    @patch('requests.get', side_effect=[requests.exceptions.ConnectionError, Mock()])
    @patch('egcg_core.integration_testing.sleep')
    @patch('egcg_core.integration_testing.get_cfg')
    @patch('egcg_core.integration_testing.check_output')
    def test_reporting_app_integration_test(self, mocked_check_output, mocked_get_cfg, mocked_sleep, mocked_get):
        # Change directory to local test folder
        test_dir = os.path.abspath(os.path.dirname(__file__))
        current_dir = os.getcwd()
        os.chdir(test_dir)
        reporting_app_data_dir = os.path.join(test_dir, 'reporting_app_data')
        if os.path.isdir(reporting_app_data_dir):
            # remove previously created directory
            shutil.rmtree(reporting_app_data_dir)
        mocked_check_output.side_effect = [
            b'docker_id\n',
            (
                b'[\n'
                b'    {\n'
                b'        "NetworkSettings": {"Networks": {"bridge": {"IPAddress": "1.2.3.4"}}},\n'
                b'        "Config": {"ExposedPorts": {"80/tcp": {}}}\n'
                b'    }\n'
                b']\n'
            ),
            b'docker_id\n',
            b'docker_id\n'
        ]
        fake_cfg = config.Configuration()
        fake_cfg.content = {
            'reporting_app': {
                'image_name': 'an_image', 'username': 'a_user', 'password': 'a_password',
                'lims_data_yaml': self.etc_config,
                'users_sqlite': self.etc_config,
                'mongo_db': self.etc
            }
        }
        mocked_get_cfg.return_value = fake_cfg
        t = integration_testing.ReportingAppIntegrationTest()
        t.setUp()

        # Check that the config file has been copied
        assert sorted(os.listdir(reporting_app_data_dir)) == sorted(['data_for_clarity_lims.yaml', 'db', 'users.sqlite'])
        assert filecmp.cmp(os.path.join(reporting_app_data_dir, 'data_for_clarity_lims.yaml'), self.etc_config)
        assert filecmp.cmp(os.path.join(reporting_app_data_dir, 'users.sqlite'), self.etc_config)
        assert sorted(os.listdir(os.path.join(reporting_app_data_dir, 'db'))) == sorted(os.listdir(self.etc))
        # Check that the command were call properly
        assert mocked_check_output.call_count == 2
        mocked_check_output.assert_any_call(
            ['docker', 'run', '-d', '-v', '%s:/opt/etc' % reporting_app_data_dir, 'an_image', 'master']
        )
        mocked_check_output.assert_any_call(['docker', 'inspect', 'docker_id'])
        # Check that the retrieval of the IP works
        mocked_get.assert_called_with('http://1.2.3.4:80/api/0.1', timeout=2)
        assert mocked_get.call_count == 2
        assert mocked_sleep.call_count == 1
        t.tearDown()
        # revert to the previous directory
        os.chdir(current_dir)



import warnings

import jinja2
import smtplib
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from email.utils import COMMASPACE
from time import sleep
from os.path import basename

from egcg_core.app_logging import AppLogger
from egcg_core.exceptions import EGCGError
from .notification import Notification


class EmailSender(AppLogger):

    def __init__(self, subject, mailhost, port, sender, recipients, email_template=None):
        self.subject = subject
        self.mailhost = mailhost
        self.port = port
        self.sender = sender
        self.recipients = recipients
        self.email_template = email_template

    def send_email(self, **kwargs):
        email = self._build_email(**kwargs)
        success = self._try_send(email)
        if not success:
            err_msg = 'Failed to send message with following args: ' + str(kwargs)
            raise EGCGError(err_msg)

    def _try_send(self, msg, retries=3):
        """
        Attempt to send an email a set number of times.
        :param int retries: Which retry we're currently on
        :return: True if a message is sucessfully sent, otherwise False
        """
        try:
            self._connect_and_send(msg)
            return True
        except OSError as e:
            retries -= 1
            self.warning('Encountered a %s exception. %s retries remaining', str(e), retries)
            if retries:
                sleep(2)
                return self._try_send(msg, retries)

            return False

    def _build_email(self, **kwargs):
        """
        Create a MIMEText from plain text or Jinja2-formatted html and send by email. If attachments are provided, the
        message will be a MIMEMultipart containing the MimeText message plus MIMEApplication attachments.

        _build_email has two modes: plain text and html. The following keyword args are useable for both:
          - email_subject: (str) override the EmailSender subject
          - email_sender: (str) override the EmailSender sender
          - email_recipients: (list) override the EmailSender recipients
          - attachments: list of file paths to attach to the email

        In plain text mode:
          - text_message (str) is required and contains the plain text to send in the email
        In html mode:
          - email_template (str) can be user to override the EmailSender email_template
          - all other keyword args are passed to the Jinja template
        """
        email_template = kwargs.get('email_template', self.email_template)
        if email_template:
            content = jinja2.Template(open(email_template).read())
            text = MIMEText(content.render(**kwargs), 'html')
        elif 'text_message' in kwargs:
            text = MIMEText(kwargs.get('text_message'))
        else:
            raise EGCGError('EmailSender needs either a text_message or an email template')
        if 'attachments' in kwargs and kwargs['attachments']:
            if isinstance(kwargs['attachments'], str):
                kwargs['attachments'] = [kwargs['attachments']]
            msg = MIMEMultipart()
            msg.attach(text)
            for attachment in kwargs['attachments']:
                with open(attachment, 'rb') as f:
                    part = MIMEApplication(f.read(), Name=basename(attachment))
                    part['Content-Disposition'] = 'attachment; filename="%s"' % basename(attachment)
                    msg.attach(part)
        else:
            msg = text

        msg['Subject'] = kwargs.get('email_subject', self.subject)
        msg['From'] = kwargs.get('email_sender', self.sender)
        msg['To'] = COMMASPACE.join(kwargs.get('email_recipients', self.recipients))
        return msg

    def _connect_and_send(self, msg):
        connection = smtplib.SMTP(self.mailhost, self.port)
        connection.send_message(msg, self.sender, self.recipients)
        connection.quit()


class EmailNotification(Notification):
    translation_map = {' ': '&nbsp', '\n': '<br/>'}

    def __init__(self, name, mailhost, port, sender, recipients, strict=False, email_template=None):
        super().__init__(name)
        self.email_sender = EmailSender(name, mailhost, port, sender, recipients, email_template)
        self.strict = strict
        self.email_template = email_template

    def notify(self, msg, attachments=None):
        try:
            if self.email_template:
                self.email_sender.send_email(title=self.name, body=self._prepare_string(msg), attachments=attachments)
            else:
                self.email_sender.send_email(text_message=msg, attachments=attachments)
        except EGCGError:
            err_msg = 'Failed to send message: ' + str(msg)
            if self.strict:
                raise EGCGError(err_msg)
            else:
                self.critical(err_msg)

    @classmethod
    def _prepare_string(cls, msg):
        for k in cls.translation_map:
            msg = msg.replace(k, cls.translation_map[k])
        return msg


def send_email(msg, mailhost, port, sender, recipients, subject, email_template=None, attachments=None, **kwargs):
    warnings.warn('send_email is deprecated - use send_plain_text_email or send_html_email instead', DeprecationWarning)
    if msg and not email_template:
        send_plain_text_email(msg, mailhost, port, sender, recipients, subject, attachments)
    elif email_template and not msg:
        send_html_email(mailhost, port, sender, recipients, subject, email_template, attachments, **kwargs)
    else:
        send_html_email(mailhost, port, sender, recipients, subject, email_template, attachments, msg=msg, **kwargs)


def send_plain_text_email(msg, mailhost, port, sender, recipients, subject, attachments=None):
    EmailSender(
        subject,
        mailhost,
        port,
        sender,
        recipients,
    ).send_email(text_message=msg, attachments=attachments)


def send_html_email(mailhost, port, sender, recipients, subject, email_template, attachments=None, **kwargs):
    EmailSender(
        subject,
        mailhost,
        port,
        sender,
        recipients,
        email_template=email_template
    ).send_email(attachments=attachments, **kwargs)



from os import makedirs
from os.path import join
import shutil
from egcg_core.executor import script_writers
from egcg_core.exceptions import EGCGError
from tests import TestEGCG

working_dir = join(TestEGCG.assets_path, 'test_script_writer_wd')


class TestScriptWriter(TestEGCG):
    writer_cls = script_writers.ScriptWriter
    array_index = 'JOB_INDEX'
    exp_header = [
        '#!/bin/bash\n',
        '# cpus: 1',
        '# exclusive',
        '# job name: a_job_name',
        '# queue: a_job_queue',
        '# job array: 1-3',
        '# log file: ' + join(working_dir, 'a_job_name.log'),
        '# mem: 2gb',
        '# walltime: 3',
        '',
        'cd ' + working_dir
    ]

    exp_minimal = [
        '#!/bin/bash\n',
        '# job name: a_job_name',
        '# queue: a_job_queue',
        '# log file: ' + join(working_dir, 'a_job_name.log'),
        '',
        'cd ' + working_dir,
        '',
        'ls'
    ]

    def setUp(self):
        self.exp_cmds = [
            '',
            'some',
            'preliminary',
            'cmds',
            'case $%s in' % self.array_index,
            '1) this\n;;',
            '2) that\n;;',
            '3) other\n;;',
            '*) echo "Unexpected %s: $%s"' % (self.array_index, self.array_index),
            'esac'
        ]

        makedirs(working_dir, exist_ok=True)
        self.script_writer = self.writer_cls(
            'a_job_name',
            working_dir,
            'a_job_queue',
            cpus=1,
            mem=2,
            walltime='3',
            exclusive=True
        )
        assert self.script_writer.lines == []

    def tearDown(self):
        shutil.rmtree(working_dir)

    def test_init(self):
        w = self.script_writer
        assert w.script_name == join(working_dir, 'a_job_name') + w.suffix
        assert w.log_commands is True
        assert w.log_file == join(working_dir, 'a_job_name.log')

    def test_register_cmd(self):
        self.script_writer.register_cmd('a_cmd', log_file='a_log_file')
        assert self.script_writer.lines == ['a_cmd > a_log_file 2>&1']

    def test_register_cmds(self):
        self.script_writer.register_cmds('this', 'that', parallel=False)
        assert self.script_writer.lines == ['this', 'that']

    def test_add_job_array(self):
        self.script_writer.add_job_array('this', 'that', 'other')
        assert self.script_writer.lines == [
            'case $%s in' % self.array_index,
            '1) this > ' + join(working_dir, 'a_job_name.log1') + ' 2>&1\n;;',
            '2) that > ' + join(working_dir, 'a_job_name.log2') + ' 2>&1\n;;',
            '3) other > ' + join(working_dir, 'a_job_name.log3') + ' 2>&1\n;;',
            '*) echo "Unexpected %s: $%s"' % (self.array_index, self.array_index),
            'esac'
        ]
        with self.assertRaises(EGCGError):
            self.script_writer.add_job_array('another', 'more')

    def test_save(self):
        self.script_writer.add_line('a_line')
        self.script_writer.save()
        assert 'a_line\n' in open(self.script_writer.script_name, 'r').readlines()

    def test(self):
        self.script_writer.log_commands = False
        self.script_writer.register_cmds('some', 'preliminary', 'cmds', parallel=False)
        self.script_writer.register_cmds('this', 'that', 'other', parallel=True)
        self.script_writer.add_header()

        obs = self.script_writer.lines
        exp = self.exp_header + self.exp_cmds
        assert obs == exp

    def test_minimal(self):
        s = self.writer_cls('a_job_name', working_dir)
        s.register_cmd('ls')
        s.add_header()
        assert s.lines == self.exp_minimal


class TestSlurmWriter(TestScriptWriter):
    writer_cls = script_writers.SlurmWriter
    array_index = 'SLURM_ARRAY_TASK_ID'
    exp_header = [
        '#!/bin/bash\n',
        '#SBATCH --cpus-per-task=1',
        '#SBATCH --exclusive',
        '#SBATCH --job-name="a_job_name"',
        '#SBATCH --partition=a_job_queue',
        '#SBATCH --array=1-3',
        '#SBATCH --output=' + join(working_dir, 'a_job_name.log'),
        '#SBATCH --mem=2g',
        '#SBATCH --time=3:00:00',
        '',
        'cd ' + working_dir
    ]

    exp_minimal = [
        '#!/bin/bash\n',
        '#SBATCH --job-name="a_job_name"',
        '#SBATCH --partition=a_job_queue',
        '#SBATCH --output=' + join(working_dir, 'a_job_name.log'),
        '',
        'cd ' + working_dir,
        '',
        'ls'
    ]



from egcg_core import constants as cst


def test_constants_uniqueness():
    constants = [item for item in dir(cst) if not item.startswith("__")]
    uniq_constants = set()
    for constant in constants:
        assert constant not in uniq_constants
        uniq_constants.add(constant)



import os
from unittest.mock import patch, Mock
from egcg_core import clarity
from egcg_core.clarity import lims_samples_info
from tests import TestEGCG


def patched(path, **kwargs):
    return patch('egcg_core.clarity.' + path, **kwargs)


def patched_lims(method, return_value=None, side_effect=None):
    return patched('_lims.' + method, return_value=return_value, side_effect=side_effect)


def patched_clarity(func, return_value=None, side_effect=None):
    return patched(func, return_value=return_value, side_effect=side_effect)


class FakeEntity(Mock):
    def __init__(self, name, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name


class FakeContainer:
    @staticmethod
    def get_placements():
        return {
            'this': Mock(samples=[FakeEntity('a name')]),
            'that': Mock(samples=[FakeEntity('another name')])
        }


class FakeProcess(Mock):
    date_run = 'a_date'
    udf = {}

    @staticmethod
    def all_inputs():
        return [Mock(samples=[FakeEntity('this'), FakeEntity('that')])]

    @staticmethod
    def input_per_sample(sample_name):
        return [Mock(id=sample_name, location=('container', '1:this'), udf={})]

    @staticmethod
    def outputs_per_input(artifact_id, **kwargs):
        return [Mock(container=artifact_id)]


fake_samples = [
    Mock(project=FakeEntity('this'), udf={'Species': 'a_species'}),
    Mock(project=FakeEntity('this'), udf={'Species': 'a_species'}),
    Mock(project=FakeEntity('that'), udf={'Species': 'a_species'})
]


class TestClarity(TestEGCG):
    def setUp(self):
        clarity._lims = Mock()
        clarity.app_logger = Mock()
        clarity._lims_samples_info.clear()

    def test_connection(self):
        clarity._lims = None
        # create connection from config
        lims = clarity.connection()
        assert lims.baseuri == 'a_baseuri/'

        # Check caching
        assert lims is clarity.connection()

        # override caching
        assert lims != clarity.connection(new=True)

        # create connection from config and param
        lims = clarity.connection(new=True, baseuri='second_uri')
        assert lims.baseuri == 'second_uri/'

    @patch('egcg_core.rest_communication.get_document', return_value={'sample_id': 'a_sample'})
    def test_lims_samples_info(self, mock_get_doc):
        assert lims_samples_info('a_sample') == {'sample_id': 'a_sample'}
        assert len(clarity._lims_samples_info) == 1
        assert clarity._lims_samples_info.get('a_sample') == {'sample_id': 'a_sample'}
        mock_get_doc.assert_called_once_with('lims/sample_info', match={'sample_id': 'a_sample'})

    def test_get_valid_lanes(self):
        fake_flowcell = Mock(
            placements={
                '1:this': Mock(udf={'Lane Failed?': False}),
                '2:that': Mock(udf={'Lane Failed?': False}),
                '3:other': Mock(udf={'Lane Failed?': True})
            }
        )
        with patched_lims('get_containers', [fake_flowcell]) as mocked_lims:
            valid_lanes = clarity.get_valid_lanes('a_flowcell_name')
            mocked_lims.assert_called_with(type='Patterned Flowcell', name='a_flowcell_name')
            assert valid_lanes == [1, 2]

    def test_find_project_from_sample(self):
        with patched_clarity('get_samples', fake_samples) as mocked_get_samples:
            project_name = clarity.find_project_name_from_sample('a_sample')
            mocked_get_samples.assert_called_with('a_sample')
            assert project_name is None
            clarity.app_logger.error.assert_called_with('%s projects found for sample %s', 2, 'a_sample')

        with patched_clarity('get_samples', fake_samples[0:1]):
            assert clarity.find_project_name_from_sample('a_sample') == 'this'

    @patched_lims('get_artifacts', [Mock(parent_process=FakeProcess)])
    @patched_clarity('get_sample', FakeEntity('a_sample'))
    def test_find_run_elements_from_sample(self, mocked_get_sample, mocked_get_artifacts):
        assert list(clarity.find_run_elements_from_sample('a_sample')) == [(None, '1')]
        mocked_get_sample.assert_called_with('a_sample')
        mocked_get_artifacts.assert_called_with(sample_name='a_sample', process_type='AUTOMATED - Sequence')

    @patched_clarity('get_species_name', 'Genus species')
    @patched_clarity('get_samples', fake_samples)
    def test_get_species_from_sample(self, mocked_get_samples, mocked_ncbi):
        clarity._lims_samples_info['a_sample_name'] = {'Species': 'genus species'}
        assert clarity.get_species_from_sample('a_sample_name') == 'Genus species'
        mocked_ncbi.assert_called_with('genus species')
        clarity._lims_samples_info['a_sample_name'] = {}

        assert clarity.get_species_from_sample('a_sample_name') == 'Genus species'
        mocked_get_samples.assert_called_with('a_sample_name')
        mocked_ncbi.assert_called_with('a_species')

    def test_get_genome_version(self):
        clarity._lims_samples_info['a_sample_name'] = {'Genome Version': 'hh45'}
        assert clarity.get_genome_version('a_sample_name', species='Homo habilis') == 'hh45'
        clarity._lims_samples_info['a_sample_name'] = {}

        with patched_clarity('get_sample', Mock(udf={'Genome Version': 'hh44'})):
            assert clarity.get_genome_version('a_sample_name', species='Homo habilis') == 'hh44'

        with patched_clarity('get_sample', Mock(udf={})), \
             patch('egcg_core.rest_communication.get_document', return_value={'default_version': 'hh43'}):
            assert clarity.get_genome_version('a_sample_name', species='Homo habilis') == 'hh43'

    def test_sanitize_user_id(self):
        assert clarity.sanitize_user_id('this?that$other another:more') == 'this_that_other_another_more'
        assert clarity.sanitize_user_id('this.that$other another:more') == 'this_that_other_another_more'

    def test_get_list_of_samples(self):
        exp_lims_sample_ids = ['this', 'that:01', 'other _L:01']
        calling_sample_ids = ['this', 'that_01', 'other__L_01']
        fake_list_samples = [[FakeEntity(n)] for n in exp_lims_sample_ids]
        psamples = patched_lims('get_samples', side_effect=fake_list_samples)

        with patched_lims('get_batch'), psamples as mocked_get_samples:
            samples = clarity.get_list_of_samples(calling_sample_ids)
            assert [s.name for s in samples] == exp_lims_sample_ids
            mocked_get_samples.assert_any_call(name=['this', 'that_01', 'other__L_01'])
            mocked_get_samples.assert_any_call(name=['other__L:01', 'that:01'])
            mocked_get_samples.assert_any_call(name=['other _L:01'])

    def test_get_list_of_samples_broken(self):
        exp_lims_sample_ids = ['this', 'that:01', 'other _L:01']
        calling_sample_ids = ['this', 'that_01', 'other__L_01']
        fake_list_samples = [[FakeEntity(n)] for n in exp_lims_sample_ids]
        psamples = patched_lims('get_samples', side_effect=fake_list_samples)
        pwarn = patched('app_logger.warning')

        with patched_lims('get_batch'), psamples as mocked_get_samples, pwarn as mocked_warn:
            samples = clarity.get_list_of_samples(calling_sample_ids + ['sample_not_in_lims'])
            assert [s.name for s in samples] == exp_lims_sample_ids
            mocked_get_samples.assert_any_call(name=['this', 'that_01', 'other__L_01', 'sample_not_in_lims'])
            mocked_get_samples.assert_any_call(name=['other__L:01', 'sample_not_in_lims', 'that:01'])
            mocked_get_samples.assert_any_call(name=['other _L:01', 'sample_not_in_lims'])
            mocked_warn.assert_called_with("Could not find %s in Lims", ['sample_not_in_lims'])

    @patched_lims('get_samples', side_effect=[[], [], [None]])
    def test_get_samples(self, mocked_lims):
        assert clarity.get_samples('a_sample_name__L_01') == [None]
        mocked_lims.assert_any_call(name='a_sample_name__L_01')
        mocked_lims.assert_any_call(name='a_sample_name__L:01')
        mocked_lims.assert_any_call(name='a_sample_name _L:01')

    @patched_clarity('get_samples', return_value=['a sample'])
    def test_get_sample(self, mocked_lims):
        assert clarity.get_sample('a_sample_id') == 'a sample'
        mocked_lims.assert_called_with('a_sample_id')

    @patched_clarity('get_sample', return_value=Mock(udf={'User Sample Name': 'a:user:sample:id'}))
    def test_get_user_sample_name(self, mocked_lims):
        clarity._lims_samples_info['a_sample_id'] = {'User Sample Name': 'uid_sample1'}
        assert clarity.get_user_sample_name('a_sample_id') == 'uid_sample1'
        clarity._lims_samples_info['a_sample_id'] = {}

        assert clarity.get_user_sample_name('a_sample_id') == 'a_user_sample_id'
        mocked_lims.assert_called_with('a_sample_id')

    @patched_clarity('get_sample', return_value=Mock(udf={'Sex': 'unknown'}))
    def test_get_sample_sex(self, mocked_lims):
        clarity._lims_samples_info['a_sample_id'] = {'Sex': 'Male'}
        assert clarity.get_sample_sex('a_sample_id') == 'Male'
        clarity._lims_samples_info['a_sample_id'] = {}

        assert clarity.get_sample_sex('a_sample_id') == 'unknown'
        mocked_lims.assert_called_with('a_sample_id')

    @patched_lims('get_file_contents', 'some test content')
    @patched_clarity('get_sample', Mock(udf={'Genotyping results file id': 1337}))
    def test_get_genotype_information_from_lims(self, mocked_get_sample, mocked_file_contents):
        genotype_vcf = os.path.join(TestEGCG.assets_path, 'a_genotype.vcf')
        assert clarity.get_sample_genotype('a_sample_name', genotype_vcf) == genotype_vcf
        mocked_get_sample.assert_called_with('a_sample_name')
        mocked_file_contents.assert_called_with(id=1337)
        assert open(genotype_vcf).read() == 'some test content'
        os.remove(genotype_vcf)

    @patched_lims('get_processes', ['a_run'])
    def test_get_run(self, mocked_lims):
        assert clarity.get_run('a_run_id') == 'a_run'
        mocked_lims.assert_called_with(type='AUTOMATED - Sequence', udf={'RunID': 'a_run_id'})

    @patched_lims('route_artifacts')
    @patched_clarity('get_list_of_samples', return_value=[Mock(artifact='this'), Mock(artifact='that')])
    @patched_lims('get_workflows', return_value=[Mock(uri='workflow_uri', stages=[Mock(uri='stage_uri')])])
    def test_route_samples_to_delivery_workflow_no_name(self, mocked_get_workflow, mocked_get_list_of_sample, mocked_route):
        clarity.route_samples_to_delivery_workflow(['a_sample_id', 'another_sample_id'])
        mocked_get_workflow.assert_called_with(name='Data Release 1.0')
        mocked_get_list_of_sample.assert_called_with(['a_sample_id', 'another_sample_id'])
        mocked_route.assert_called_with(['this', 'that'], stage_uri='stage_uri')

    @patched_lims('route_artifacts')
    @patched_clarity('get_list_of_samples', return_value=[Mock(artifact='this'), Mock(artifact='that')])
    @patched_lims('get_workflows', return_value=[Mock(uri='workflow_uri', stages=[Mock(uri='stage_uri')])])
    def test_route_samples_to_delivery_workflow_with_name(self, mocked_get_workflow, mocked_get_list_of_sample, mocked_route):
        clarity.route_samples_to_delivery_workflow(['a_sample_id', 'another_sample_id'], workflow_name='Much better workflow')
        mocked_get_workflow.assert_called_with(name='Much better workflow')
        mocked_get_list_of_sample.assert_called_with(['a_sample_id', 'another_sample_id'])
        mocked_route.assert_called_with(['this', 'that'], stage_uri='stage_uri')

    @patched_clarity('get_samples', [Mock(artifact=Mock(location=(FakeEntity('a_plate'), 'a_well')))])
    def test_get_plate_id_and_well_from_lims(self, mocked_lims):
        assert clarity.get_plate_id_and_well('a_sample_id') == ('a_plate', 'a_well')
        mocked_lims.assert_called_with('a_sample_id')

    @patched_lims('get_containers', [FakeContainer])
    def test_get_sample_names_from_plate_from_lims(self, mocked_lims):
        obs = clarity.get_sample_names_from_plate('a_plate_id')
        assert sorted(obs) == ['a_name', 'another_name']
        mocked_lims.assert_called_with(type='96 well plate', name='a_plate_id')

    @patched_lims('get_samples', [FakeEntity('this'), FakeEntity('that')])
    def test_get_sample_names_from_project_from_lims(self, mocked_lims):
        assert clarity.get_sample_names_from_project('a_project') == ['this', 'that']
        mocked_lims.assert_called_with(projectname='a_project')

    @patched_lims('get_processes', [FakeProcess])
    @patched_lims('get_artifacts', [Mock(id='this'), Mock(id='that')])
    @patched_clarity('get_sample', FakeEntity('a_sample_name'))
    def test_get_output_containers_from_sample_and_step_name(self, mocked_get_sample, mocked_get_arts, mocked_get_prcs):
        obs = clarity.get_output_containers_from_sample_and_step_name('a_sample_id', 'a_step_name')
        assert obs == {'a_sample_name'}
        mocked_get_sample.assert_called_with('a_sample_id')
        mocked_get_arts.assert_called_with(sample_name='a_sample_name')
        mocked_get_prcs.assert_called_with(type='a_step_name', inputartifactlimsid=['this', 'that'])

    @patched_clarity('get_sample_names_from_plate', ['this', 'that', 'other'])
    @patched_clarity('get_sample', Mock(artifact=Mock(container=FakeEntity('a_container', type=FakeEntity('96 well plate')))))
    def test_get_samples_arrived_with(self, mocked_get_sample, mocked_names_from_plate):
        assert clarity.get_samples_arrived_with('a_sample_name') == ['this', 'that', 'other']
        mocked_get_sample.assert_called_with('a_sample_name')
        mocked_names_from_plate.assert_called_with('a_container')

    @patched_clarity('get_sample_names_from_plate', ['other'])
    @patched_clarity('get_output_containers_from_sample_and_step_name', [FakeEntity('this'), FakeEntity('that')])
    @patched_clarity('get_sample', FakeEntity('a_sample_name'))
    def test_get_samples_genotyped_with(self, mocked_get_sample, mocked_containers, mocked_names_from_plate):
        assert clarity.get_samples_genotyped_with('a_sample_name') == {'other'}
        mocked_get_sample.assert_called_with('a_sample_name')
        mocked_containers.assert_called_with('a_sample_name', 'Genotyping Plate Preparation EG 1.0')
        mocked_names_from_plate.assert_any_call('this')
        mocked_names_from_plate.assert_any_call('that')

    @patched_clarity('get_sample_names_from_plate', ['other'])
    @patched_clarity('get_output_containers_from_sample_and_step_name', [FakeEntity('this'), FakeEntity('that')])
    @patched_clarity('get_sample', FakeEntity('a_sample_name'))
    def test_get_samples_sequenced_with(self, mocked_get_sample, mocked_containers, mocked_names_from_plate):
        assert clarity.get_samples_sequenced_with('a_sample_name') == {'other'}
        mocked_get_sample.assert_called_with('a_sample_name')
        mocked_containers.assert_called_with('a_sample_name', 'Sequencing Plate Preparation EG 1.0')
        mocked_names_from_plate.assert_any_call('this')
        mocked_names_from_plate.assert_any_call('that')

    @patched_clarity('get_sample', Mock(artifact=Mock(id='an_artifact_id')))
    @patched_lims('get_processes', side_effect=[[FakeProcess], [], [FakeProcess, FakeProcess(date_run='another_date')],
                                                [], [], []])
    def test_get_sample_release_date(self, mocked_get_procs, mocked_get_sample):
        assert clarity.get_sample_release_date('a_sample_name') == 'a_date'
        mocked_get_procs.assert_any_call(type='Data Release EG 1.0', inputartifactlimsid='an_artifact_id')
        mocked_get_procs.assert_any_call(type='Data Release EG 2.0', inputartifactlimsid='an_artifact_id')
        assert mocked_get_procs.call_count == 2
        mocked_get_sample.assert_called_with('a_sample_name')
        mocked_get_procs.reset_mock()
        mocked_get_sample.reset_mock()

        assert clarity.get_sample_release_date('a_sample_name2') == 'another_date'
        clarity.app_logger.warning.assert_called_with(
            '%s Processes found for sample %s: returning latest one', 2, 'a_sample_name2'
        )
        mocked_get_procs.reset_mock()
        mocked_get_sample.reset_mock()

        # get_processes return empty lists
        assert clarity.get_sample_release_date('a_sample_name2') is None

        # Sample does not exist
        mocked_get_sample.return_value = None
        mocked_get_procs.reset_mock()
        mocked_get_sample.reset_mock()
        assert clarity.get_sample_release_date('a_sample_name2') is None




import subprocess
from egcg_core.app_logging import AppLogger
from egcg_core.exceptions import EGCGError


class Executor(AppLogger):
    def __init__(self, cmd):
        self.cmd = cmd
        self.proc = None

    def join(self):
        """
        Set self.proc to a Popen and start.
        :rtype: tuple[bytes, bytes]
        :raises: EGCGError on any exception
        """
        try:
            out, err = self._process().communicate()
            for stream, emit in ((out, self.info), (err, self.error)):
                for line in stream.decode('utf-8').split('\n'):
                    emit(line)
            return self.proc.poll()

        except Exception as e:
            raise EGCGError('Command failed: ' + self.cmd) from e

    @staticmethod
    def start():
        """
        Subclasses of Executor implement this method from different sources:
          - StreamExecutor and ArrayExecutor inherit it from threading.Thread
          - ClusterExecutor implements it so it writes and submits a script to a queue
        The definition here has no effect, but allows Executor to be handled identically to other executors.
        """
        pass

    def _process(self):
        """
        Translate self.cmd to a subprocess. Override to manipulate how the process is run, e.g. with different
        resource managers.
        :rtype: subprocess.Popen
        """
        self.info('Executing: %s', self.cmd)
        self.proc = subprocess.Popen(self.cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        return self.proc



#!/usr/bin/env python
import os
import sys
import shutil
import random
import pytest
import argparse
from io import StringIO
from string import hexdigits
from subprocess import check_call, check_output, CalledProcessError
from contextlib import redirect_stdout
from egcg_core import notifications, integration_testing


def resolve_path(fp):
    if fp and not os.path.isabs(fp):
        return os.path.join(os.getcwd(), fp)
    return fp


def random_string(strlen=6):
    s = ''
    for i in range(strlen):
        s += random.choice(hexdigits)
    return s


def main():
    a = argparse.ArgumentParser()
    g = a.add_mutually_exclusive_group(required=True)
    g.add_argument('--local', help='Use a copy of a local directory for testing - must be an absolute path')
    g.add_argument('--remote', help='Use a checked out Git repo for testing')
    a.add_argument('--branch', help='Checkout a branch in the working dir')
    a.add_argument('--integration_cfg', help='Absolute path to an integration test config file')
    a.add_argument('--integration_test_target', default='integration_tests', help='Subdir in the working dir containing the integration tests')
    a.add_argument('--app_config')
    a.add_argument('--app_config_env_var')
    a.add_argument('--coverage_targets', nargs='+', help='Submodules to pass to pytest-cov for coverage metrics')
    a.add_argument('--stdout', action='store_true', help='Write test results to stdout')
    a.add_argument('--email', action='store_true', help='Email test results')
    a.add_argument('--log_repo', help='Log test results to a timestamped file in log_repo')
    a.add_argument('--test', nargs='+', default=[], help='Whitelist of integration tests to run. Run everything if not specified')
    a.add_argument('--ls', action='store_true', help='List all available integration tests')
    a.add_argument('--nocleanup', dest='cleanup', action='store_false', help='Don\'t clean up working dir afterwards')
    a.add_argument('-n', help='Number of concurrent processes to run tests in')
    args = a.parse_args()

    top_level = os.getcwd()

    app_config_master_copy = resolve_path(args.app_config)
    if args.integration_cfg:
        os.environ['INTEGRATIONCONFIG'] = args.integration_cfg

    run_dir = os.path.join(top_level, 'integration_test_%s' % random_string())
    os.mkdir(run_dir)
    os.chdir(run_dir)

    if args.local:
        assert os.path.isabs(args.local)
        checked_out_project = os.path.basename(args.local)
        assert os.path.isabs(args.local), 'Need an absolute path for local dirs'
        check_call(['cp', '-r', args.local, checked_out_project])  # shutil copytree fails for some reason
    else:
        assert args.remote
        check_call(['git', 'clone', args.remote])
        checked_out_project = os.path.basename(args.remote).split('.')[0]

    assert os.path.isdir(checked_out_project)
    os.chdir(checked_out_project)

    if args.branch:
        check_call(['git', 'checkout', args.branch])

    if app_config_master_copy:
        app_config = os.path.join(os.getcwd(), 'app_config.yaml')
        shutil.copy(app_config_master_copy, app_config)
        if args.app_config_env_var:
            os.environ[args.app_config_env_var] = app_config

    sys.path.append(os.getcwd())
    pytest_args = [args.integration_test_target, '--tb=short']

    if args.ls:
        pytest_args.append('--collect-only')
    else:
        if args.test:
            pytest_args.extend(['-k', ' or '.join(args.test)])

        if args.coverage_targets:
            pytest_args.append('--cov')
            pytest_args.extend(args.coverage_targets)
            pytest_args.extend(['--cov-report', 'term-missing'])

        if args.n:
            pytest_args.extend(['-n', args.n])

    checks_log = 'checks.log'
    start_time = integration_testing.now()
    with open(checks_log, 'w') as f:
        f.write('test_method\tcheck_name\tassert_method\tresult\targs\n')

    s = StringIO()
    with redirect_stdout(s):
        exit_status = pytest.main(pytest_args)
    end_time = integration_testing.now()

    try:
        git_commit = check_output(['git', 'log', '--format=Run on commit %h%d, made on %aD', '-1']).decode()
    except CalledProcessError:
        git_commit = 'Git info not available'

    test_output = '------\nIntegration test finished\n%s\nStart time: %s, finish time: %s\nPytest output:\n' % (
        git_commit, start_time, end_time
    )

    test_output += s.getvalue()
    if os.path.isfile(checks_log):
        with open(checks_log, 'r') as f:
            checks = f.read()
            if checks:
                test_output += 'Checks:\n' + checks

    os.chdir(top_level)

    if args.log_repo:
        with open(os.path.join(args.log_repo, start_time + '.log'), 'w') as f:
            f.write(test_output)

    if args.stdout:
        print(test_output)

    if args.email:
        notifications.send_plain_text_email(
            test_output,
            subject=checked_out_project + ' integration test',
            **integration_testing.get_cfg()['notification']
        )

    if args.cleanup:
        shutil.rmtree(run_dir)

    return exit_status


if __name__ == '__main__':
    main()



class EGCGError(Exception):
    pass


class ConfigError(EGCGError):
    pass


class RestCommunicationError(EGCGError):
    pass


class LimsCommunicationError(EGCGError):
    pass


class ArchivingError(EGCGError):
    pass



import hashlib
from os import makedirs, symlink
from shutil import rmtree
from os.path import join, basename
from unittest.mock import patch
from egcg_core import util
from tests import TestEGCG

fastq_dir = join(TestEGCG.assets_path, 'fastqs')


def test_find_files():
    expected = [join(TestEGCG.assets_path, f) for f in ('ftest.txt', 'ftest_2.txt')]
    assert util.find_files(TestEGCG.assets_path, 'ftest*.txt') == expected


@patch('logging.Logger.warning')
def test_find_file(mocked_log):
    assert util.find_file(TestEGCG.assets_path, 'ftest.txt') == join(TestEGCG.assets_path, 'ftest.txt')
    assert util.find_file(TestEGCG.assets_path, 'ftest_.txt') is None
    assert util.find_file(TestEGCG.assets_path, 'ftest*.txt') is None
    mocked_log.assert_called_with(
        'Searched pattern %s for one file, but got %s', (TestEGCG.assets_path, 'ftest*.txt'), 2
    )


def test_str_join():
    assert util.str_join('this', 'that', 'other', separator='/') == 'this/that/other'


def test_find_fastqs():
    fastqs = util.find_fastqs(fastq_dir, '10015AT', '10015AT0001')
    for file_name in ('10015AT0001_S6_L004_R1_001.fastq.gz', '10015AT0001_S6_L004_R2_001.fastq.gz',
                      '10015AT0001_S6_L005_R1_001.fastq.gz', '10015AT0001_S6_L005_R2_001.fastq.gz'):
        assert join(fastq_dir, '10015AT', '10015AT0001', file_name) in fastqs


def test_find_fastqs_with_lane():
    fastqs = util.find_fastqs(fastq_dir, '10015AT', '10015AT0001', lane=4)
    for file_name in ('10015AT0001_S6_L004_R1_001.fastq.gz', '10015AT0001_S6_L004_R2_001.fastq.gz'):
        assert join(fastq_dir, '10015AT', '10015AT0001', file_name) in fastqs


def test_find_all_fastqs():
    fastqs = util.find_all_fastqs(fastq_dir)
    for file_name in ('10015AT0001_S6_L004_R1_001.fastq.gz', '10015AT0001_S6_L004_R2_001.fastq.gz',
                      '10015AT0002_merged_R1.fastq.gz', '10015AT0002_merged_R2.fastq.gz'):
        assert file_name in [basename(f) for f in fastqs]


def test_find_all_fastq_pairs():
    observed = util.find_all_fastq_pairs(join(fastq_dir, '10015AT', '10015AT0001'))
    expected = [('10015AT0001_S6_L004_R1_001.fastq.gz', '10015AT0001_S6_L004_R2_001.fastq.gz'),
                ('10015AT0001_S6_L005_R1_001.fastq.gz', '10015AT0001_S6_L005_R2_001.fastq.gz')]
    assert [(basename(f), basename(g)) for f, g in observed] == expected


def test_same_fs():
    test = join(TestEGCG.assets_path, 'ftest.txt')
    test_2 = join(TestEGCG.assets_path, 'ftest_2.txt')
    test_nonexistent = join(TestEGCG.assets_path, 'ftest_nonexistent.txt')

    assert util.same_fs(test, None) is False
    assert util.same_fs(test, test_2)
    assert util.same_fs(test, test_nonexistent)


class TestMoveDir(TestEGCG):
    @staticmethod
    def _create_test_file(f, content='This is a test file'):
        with open(f, 'w') as of:
            of.write(content)

    @staticmethod
    def _md5(f):
        hash_md5 = hashlib.md5()
        with open(f, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b''):
                hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def setUp(self):
        self.test_dir = join(self.assets_path, 'move_dir')
        makedirs(join(self.test_dir, 'from', 'subdir'), exist_ok=True)
        self._create_test_file(join(self.test_dir, 'from', 'ftest.txt'))
        self._create_test_file(join(self.test_dir, 'from', 'subdir', 'ftest.txt'))

        makedirs(join(self.test_dir, 'external'), exist_ok=True)
        self._create_test_file(join(self.test_dir, 'external', 'external.txt'), 'External file')
        symlink(join(self.test_dir, 'external', 'external.txt'), join(self.test_dir, 'from', 'external_renamed.txt'))

        makedirs(join(self.test_dir, 'exists', 'subdir'), exist_ok=True)
        self._create_test_file(join(self.test_dir, 'exists', 'subdir', 'ftest.txt'), 'another file')
        self._create_test_file(join(self.test_dir, 'exists', 'ftest.txt'), 'another file')

    def tearDown(self):
        for base in ('to', 'from', 'exists', 'external'):
            f = util.find_file(self.test_dir, base)
            if f:
                rmtree(f)

    def test_move_dir(self):
        frm = join(self.test_dir, 'from')
        to = join(self.test_dir, 'to')
        md5_from = self._md5(join(frm, 'ftest.txt'))
        assert util.find_file(frm, 'ftest.txt')
        assert not util.find_file(to)

        assert util.move_dir(frm, to) == 0

        assert not util.find_file(frm, 'ftest.txt')
        assert util.find_file(to, 'ftest.txt')
        assert util.find_file(to, 'subdir', 'ftest.txt')
        assert md5_from == self._md5(join(to, 'ftest.txt'))

        assert util.find_file(to, 'external_renamed.txt')

    def test_move_dir_exists(self):
        frm = join(self.test_dir, 'from')
        to = join(self.test_dir, 'exists')
        md5_from1 = self._md5(join(frm, 'ftest.txt'))
        md5_from2 = self._md5(join(frm, 'subdir', 'ftest.txt'))

        assert util.find_file(frm, 'ftest.txt')
        assert util.find_file(to, 'ftest.txt')
        assert not md5_from1 == self._md5(join(to, 'ftest.txt'))
        assert not md5_from2 == self._md5(join(to, 'subdir', 'ftest.txt'))

        util.move_dir(frm, to)

        assert not util.find_file(frm, 'ftest.txt')
        assert util.find_file(to, 'ftest.txt')
        assert md5_from1 == self._md5(join(to, 'ftest.txt'))
        assert md5_from2 == self._md5(join(to, 'subdir', 'ftest.txt'))


def test_query_dict():
    data = {'this': {'that': 'other'}}
    assert util.query_dict(data, 'this') == {'that': 'other'}
    assert util.query_dict(data, 'this.that') == 'other'
    assert util.query_dict(data, 'nonexistent') is None
    assert util.query_dict(data, 'nonexistent', ret_default='things') == 'things'



from threading import Thread
from egcg_core.exceptions import EGCGError
from .stream_executor import StreamExecutor


class ArrayExecutor(StreamExecutor):
    def __init__(self, cmds, stream):
        """
        :param cmds:
        :param bool stream: Whether to run all commands in parallel or one after another
        """
        super().__init__(cmds)
        self.executors = []
        self.exit_statuses = []
        self.stream = stream
        for c in cmds:
            self.executors.append(StreamExecutor(c))

    def run(self):
        try:
            if self.stream:
                for e in self.executors:
                    e.start()
                for e in self.executors:
                    self.exit_statuses.append(e.join())
            else:
                for e in self.executors:
                    e.start()
                    self.exit_statuses.append(e.join())
        except Exception as err:
            self.exception = err

    def join(self, timeout=None):
        # noinspection PyCallByClass
        Thread.join(self, timeout)
        if self.exception:
            self._stop()
            self.error(self.exception.__class__.__name__ + ': ' + str(self.exception))
            raise EGCGError('Commands failed: ' + str(self.exit_statuses)) from self.exception
        self.info('Exit statuses: %s', self.exit_statuses)
        return sum(self.exit_statuses)



import os
import shutil
from glob import glob
from egcg_core.exceptions import EGCGError
from egcg_core.app_logging import logging_default as log_cfg

app_logger = log_cfg.get_logger(__name__)


def check_if_nested(data):
    """Check whether a data structure contains lists/dicts."""
    if isinstance(data, dict):
        for k in data:
            if isinstance(data[k], (list, dict)):
                return True
    elif isinstance(data, list):
        for i in data:
            if isinstance(i, (list, dict)):
                return True
    return False


def find_files(*path_parts):
    return sorted(glob(os.path.join(*path_parts)))


def find_file(*path_parts):
    files = find_files(*path_parts)
    nfiles = len(files)
    if nfiles > 1:
        app_logger.warning('Searched pattern %s for one file, but got %s', path_parts, nfiles)
    elif nfiles == 1:
        return files[0]


def str_join(*parts, separator=''):
    return separator.join(parts)


def find_fastqs(location, project_id, sample_id, lane=None):
    """
    Find all .fastq.gz files in an input folder 'location/project_id'.
    :param location: Top-level directory
    :param str project_id: Project subdirectory to search
    :param str sample_id: Sample subdirectory to search
    :param lane: Specific lane to search for (optional)
    """
    basename = '*.fastq.gz'
    if lane:
        basename = '*L00' + str(lane) + basename

    pattern = os.path.join(location, project_id, sample_id, basename)
    fastqs = find_files(pattern)
    app_logger.debug('Found %s fastq files for %s', len(fastqs), pattern)
    return fastqs


def find_all_fastqs(location):
    """Find fastqs in an input folder, regardless of directory structure."""
    fastqs = []
    for name, dirs, files in os.walk(location):
        fastqs.extend(os.path.join(name, f) for f in files if f.endswith('.fastq.gz'))
    app_logger.debug('Found %s fastqs in %s', len(fastqs), location)
    return fastqs


def find_all_fastq_pairs(location):
    """
    Return the results of find_all_fastqs as a list of fastq pair tuples. Assumes that fastq pairs come
    together from sorting by name.
    :param str location: Directory to search
    :return: Full paths to all fastq.gz files in the input dir, aggregated per pair
    :rtype: list[tuple[str, str]]
    """
    fastqs = find_all_fastqs(location)
    if len(fastqs) % 2 != 0:
        raise EGCGError('Expected even number of fastq files in %s, found %s' % (location, len(fastqs)))
    fastqs.sort()
    return list(zip(*[iter(fastqs)] * 2))


def same_fs(file1, file2):
    if not file1 or not file2:
        return False
    if not os.path.exists(file1):
        return same_fs(os.path.dirname(file1), file2)
    if not os.path.exists(file2):
        return same_fs(file1, os.path.dirname(file2))

    dev1 = os.stat(file1).st_dev
    dev2 = os.stat(file2).st_dev
    return dev1 == dev2


def move_dir(src_dir, dest_dir):
    os.makedirs(dest_dir, exist_ok=True)
    for f in os.listdir(src_dir):
        src_file = os.path.join(src_dir, f)
        if os.path.isdir(src_file):
            move_dir(src_file, os.path.join(dest_dir, f))
        else:
            fp = os.path.realpath(src_file)
            dest_file = os.path.join(dest_dir, os.path.basename(src_file))
            shutil.move(fp, dest_file)
    return 0


def query_dict(data, query_string, ret_default=None):
    """
    Drill down into a dict using dot notation, e.g. query_dict({'this': {'that': 'other'}}, 'this.that'}).
    :param dict data:
    :param str query_string:
    :param ret_default:
    """
    _data = data.copy()

    for q in query_string.split('.'):
        d = _data.get(q)
        if d is not None:
            _data = d
        else:
            return ret_default

    return _data



from setuptools import setup, find_packages
from os.path import join, abspath, dirname
requirements_txt = join(abspath(dirname(__file__)), 'requirements.txt')
requirements = [l.strip() for l in open(requirements_txt) if l and not l.startswith('#')]


def _translate_req(r):
    # this>=0.3.2 -> this(>=0.3.2)
    ops = ('<=', '>=', '==', '<', '>', '!=')
    _version = None
    for op in ops:
        if op in r:
            r, _version = r.split(op)
            _version = op + _version

    req = r
    if _version:
        req += '(%s)' % _version
    return req

version = '0.13'

setup(
    name='EGCG-Core',
    version=version,
    packages=find_packages(exclude=('tests',)),
    url='https://github.com/EdinburghGenomics/EGCG-Core',
    license='MIT',
    description='Shared functionality across EGCG projects',
    long_description='Common modules for use across EGCG projects. Includes logging, configuration, common '
                     'exceptions, random utility functions, and modules for interfacing with external data '
                     'sources such as EGCG\'s reporting app and Clarity LIMS instance',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.6",
    ],
    keywords='EdinburghGenomics executor notification logging api rest',
    requires=[_translate_req(r) for r in requirements],  # metadata
    install_requires=requirements,  # actual module requirements
    scripts=['bin/integration_test_runner.py'],
    zip_safe=False,
    author='Murray Wham',
    author_email='murray.wham@ed.ac.uk'
)



import pytest
from unittest.mock import Mock, patch
from os import remove
from os.path import join, abspath, dirname
from smtplib import SMTPException
from email.mime.text import MIMEText
from email.mime.application import MIMEApplication
from email.mime.multipart import MIMEMultipart
from egcg_core import notifications as n
from egcg_core.exceptions import EGCGError
from tests import TestEGCG


class FakeSMTP(Mock):
    def __init__(self, host, port):
        super().__init__()
        self.mailhost, self.port = host, port

    @staticmethod
    def send_message(msg, reporter, recipients):
        if 'dodgy' in str(msg):
            raise SMTPException('Oh noes!')
        else:
            pass

    @staticmethod
    def quit():
        pass


class TestNotificationCentre(TestEGCG):
    def setUp(self):
        self.notification_centre = n.NotificationCentre('a_name')

    def test_config_init(self):
        e = self.notification_centre.subscribers['email']
        assert e.email_sender.sender == 'this'
        assert e.email_sender.recipients == ['that', 'other']
        assert e.email_sender.mailhost == 'localhost'
        assert e.email_sender.port == 1337
        assert e.strict is True

        a = self.notification_centre.subscribers['asana']
        assert a.workspace_id == 1337
        assert a.project_id == 1338

    def test_notify(self):
        self.notification_centre.subscribers = {'asana': Mock(), 'email': Mock()}
        self.notification_centre.notify_all('a message')
        for name, s in self.notification_centre.subscribers.items():
            s.notify.assert_called_with('a message')

    @patch.object(n.NotificationCentre, 'critical')
    def test_notify_failure(self, mocked_log):
        self.notification_centre.subscribers = {
            'asana': Mock(notify=Mock(side_effect=EGCGError('Something broke'))),
            'email': Mock(notify=Mock(side_effect=ValueError('Something else broke')))
        }
        with self.assertRaises(EGCGError) as e:
            self.notification_centre.notify('a message', ('asana', 'email'))

        mocked_log.assert_called()
        assert str(e.exception) == (
            'Encountered the following errors during notification: '
            'EGCGError: Something broke, '
            'ValueError: Something else broke'
        )


class TestLogNotification(TestEGCG):
    def setUp(self):
        self.notification_log = join(self.assets_path, 'LogNotification.log')
        self.ntf = n.LogNotification('log_notification', self.notification_log)

    def tearDown(self):
        remove(self.notification_log)

    def test_notify(self):
        self.ntf.notify('message')
        with open(self.notification_log) as open_file:
            assert '[log_notification] message' in open_file.read()

    def _test_notify_with_attachments(self, attachments):
        attachment = join(self.assets_path, 'test_to_upload.txt')
        self.ntf.notify('message', attachments=attachments)
        with open(self.notification_log) as open_file:
            data = open_file.read().split('\n')
            assert '[log_notification] message' in data[0]
            assert '[log_notification] Attachment: %s' % attachment in data[1]

    def test_notify_with_attachment(self):
        self._test_notify_with_attachments([join(self.assets_path, 'test_to_upload.txt')])

    def test_notify_with_attachments(self):
        self._test_notify_with_attachments(join(self.assets_path, 'test_to_upload.txt'))


class TestEmailSender(TestEGCG):
    def setUp(self):
        self.email_sender = n.EmailSender(
            'a_subject', 'localhost', 1337, 'a_sender', ['some', 'recipients'],
            email_template=join(self.etc, 'email_notification.html')
        )

    @patch('egcg_core.notifications.email.EmailSender._logger')
    def test_retries(self, mocked_logger):
        self.email_sender.email_template = None
        with patch('smtplib.SMTP', new=FakeSMTP), patch('egcg_core.notifications.email.sleep'):
            assert self.email_sender._try_send(self.email_sender._build_email(text_message='this is a test')) is True
            assert self.email_sender._try_send(self.email_sender._build_email(text_message='dodgy')) is False
            for i in range(3):
                mocked_logger.warning.assert_any_call(
                    'Encountered a %s exception. %s retries remaining', 'Oh noes!', i
                )

            with pytest.raises(EGCGError) as e:
                self.email_sender.send_email(text_message='dodgy')
                assert 'Failed to send message: dodgy' in str(e)

    def test_build_email(self):
        exp_msg = (
            '<!DOCTYPE html>\n'
            '<html lang="en">\n'
            '<head>\n'
            '    <meta charset="UTF-8">\n'
            '    <style>\n'
            '        table, th, td {\n'
            '            border: 1px solid black;\n'
            '            border-collapse: collapse;\n'
            '        }\n'
            '    </style>\n'
            '</head>\n'
            '<body>\n'
            '    <h2>a_subject</h2>\n'
            '    <p>a message</p>\n'
            '</body>\n'
            '</html>'
        )

        exp = MIMEText(exp_msg, 'html')
        exp['Subject'] = 'a_subject'
        exp['From'] = 'a_sender'
        exp['To'] = 'some, recipients'
        obs = self.email_sender._build_email(title='a_subject', body='a message')
        assert str(obs) == str(exp)

    def test_build_email_plain_text(self):
        self.email_sender.email_template = None
        exp = MIMEText('a message')
        exp['Subject'] = 'a_subject'
        exp['From'] = 'a_sender'
        exp['To'] = 'some, recipients'
        obs = self.email_sender._build_email(text_message='a message')
        assert str(obs) == str(exp)

    def _test_build_email_attachments(self, attachements):
        attachment = join(self.assets_path, 'test_to_upload.txt')
        self.email_sender.email_template = None
        exp = MIMEMultipart()
        exp['Subject'] = 'a_subject'
        exp['From'] = 'a_sender'
        exp['To'] = 'some, recipients'
        obs = self.email_sender._build_email(text_message='a message', attachments=attachements)
        payload = obs.get_payload()
        assert len(payload) == 2
        assert str(payload[0]) == str(MIMEText('a message'))
        with open(attachment, 'rb') as open_file:
            part = MIMEApplication(
                open_file.read(),
                Name='test_to_upload.txt'
            )
            part['Content-Disposition'] = 'attachment; filename="test_to_upload.txt"'
            assert str(payload[1]) == str(part)

    def test_build_email_attachments(self):
        attachment = join(self.assets_path, 'test_to_upload.txt')
        self._test_build_email_attachments([attachment])

    def test_build_email_attachment(self):
        attachment = join(self.assets_path, 'test_to_upload.txt')
        self._test_build_email_attachments(attachment)


class TestEmailNotification(TestEGCG):
    def setUp(self):
        self.ntf = n.EmailNotification('a_subject', 'localhost', 1337, 'a_sender', ['some', 'recipients'], strict=True)
        self.ntf2 = n.EmailNotification(
            'a_subject', 'localhost', 1337, 'a_sender', ['some', 'recipients'], strict=True,
            email_template=join(dirname(dirname(abspath(__file__))), 'etc', 'email_notification.html')
        )

    @patch.object(n.EmailSender, 'send_email')
    def test_notify(self, mock_send_email):
        self.ntf.notify('a message')
        mock_send_email.assert_called_once_with(text_message='a message', attachments=None)

        mock_send_email.reset_mock()
        self.ntf2.notify('a message')
        mock_send_email.assert_called_once_with(title='a_subject', body='a&nbspmessage', attachments=None)

    @patch.object(n.EmailSender, 'send_email', side_effect=EGCGError)
    def test_notify_fail(self, mock_send_email):
        with pytest.raises(EGCGError) as excinfo:
            self.ntf.notify('a message')
        assert 'Failed to send message: a message' in str(excinfo)


@patch('egcg_core.notifications.email.EmailSender._try_send')
def test_send_plain_text_email(mocked_send):
    n.send_plain_text_email('a message', 'localhost', 1337, 'a_sender', ['some', 'recipients'], 'a subject')
    exp = MIMEText('a message')
    exp['Subject'] = 'a subject'
    exp['From'] = 'a_sender'
    exp['To'] = 'some, recipients'
    assert str(mocked_send.call_args[0][0]) == str(exp)


@patch('egcg_core.notifications.email.EmailSender._try_send')
def test_send_html_email(mocked_send):
    email_template = join(dirname(dirname(abspath(__file__))), 'etc', 'email_notification.html')

    n.send_html_email('localhost', 1337, 'Sender', ['recipient1', 'recipient2'], 'Subject',
                      email_template=email_template, title='title', body='body')
    exp_msg = (
        '<!DOCTYPE html>\n'
        '<html lang="en">\n'
        '<head>\n'
        '    <meta charset="UTF-8">\n'
        '    <style>\n'
        '        table, th, td {\n'
        '            border: 1px solid black;\n'
        '            border-collapse: collapse;\n'
        '        }\n'
        '    </style>\n'
        '</head>\n'
        '<body>\n'
        '    <h2>title</h2>\n'
        '    <p>body</p>\n'
        '</body>\n'
        '</html>'
    )
    exp = MIMEText(exp_msg, 'html')
    exp['Subject'] = 'Subject'
    exp['From'] = 'Sender'
    exp['To'] = 'recipient1, recipient2'
    assert str(mocked_send.call_args[0][0]) == str(exp)


@patch('egcg_core.notifications.email.send_plain_text_email')
@patch('egcg_core.notifications.email.send_html_email')
def test_send_email(mocked_html, mocked_plain_text):
    email_template = join(dirname(dirname(abspath(__file__))), 'etc', 'email_notification.html')

    n.send_email('a message', 'localhost', 1337, 'Sender', ['recipient1', 'recipient2'], 'Subject')
    assert mocked_plain_text.call_count == 1
    assert mocked_html.call_count == 0

    n.send_email(None, 'localhost', 1337, 'Sender', ['Recipients'], 'Subject', email_template, a_jinja2='arg')
    mocked_html.assert_called_with('localhost', 1337, 'Sender', ['Recipients'], 'Subject', email_template, None, a_jinja2='arg')

    # Adding message and template means the template is used and the msg is only passed and most likely ignored
    n.send_email('msg', 'localhost', 1337, 'Sender', ['Recipients'], 'Subject', email_template, a_jinja2='arg')
    mocked_html.assert_called_with('localhost', 1337, 'Sender', ['Recipients'], 'Subject', email_template, None, a_jinja2='arg', msg='msg')


class TestAsanaNotification(TestEGCG):
    def setUp(self):
        self.ntf = n.AsanaNotification(
            'another_name',
            workspace_id=1337,
            project_id=1338,
            access_token='an_access_token'
        )
        self.ntf.client = Mock(
            tasks=Mock(
                find_all=Mock(return_value=[{'name': 'this'}]),
                create_in_workspace=Mock(return_value={'gid': 1337}),
                find_by_id=Mock(return_value={'name': 'this', 'gid': 1337})
            )
        )

    def test_task(self):
        assert self.ntf.task == {'gid': 1337, 'name': 'this'}
        self.ntf.client.tasks.find_by_id.assert_called_with(1337)

    def test_add_comment(self):
        self.ntf.notify('a comment')
        self.ntf.client.tasks.add_comment.assert_called_with(1337, text='a comment')

    def _test_add_comment_with_attachments(self, attachments):
        self.ntf.notify('a comment', attachments=attachments)
        self.ntf.client.tasks.add_comment.assert_called_with(1337, text='a comment')
        self.ntf.client.request.assert_called_with(
            'post',
            '/tasks/1337/attachments',
            files=[('file', ('test_to_upload.txt', b'test content', None))]
        )

    def test_add_comment_with_attachments(self):
        self._test_add_comment_with_attachments([join(self.assets_path, 'test_to_upload.txt')])

    def test_add_comment_with_one_attachment(self):
        self._test_add_comment_with_attachments(join(self.assets_path, 'test_to_upload.txt'))

    def test_get_entity(self):
        collection = [{'name': 'this'}, {'name': 'that'}]
        assert self.ntf._get_entity(collection, 'that') == {'name': 'that'}
        assert self.ntf._get_entity(collection, 'other') is None

    def test_create_task(self):
        assert self.ntf._create_task() == {'gid': 1337}
        self.ntf.client.tasks.create_in_workspace.assert_called_with(1337, self.ntf.task_template)



from select import select
from threading import Thread
from egcg_core.executor import Executor
from egcg_core.exceptions import EGCGError


class StreamExecutor(Thread, Executor):
    def __init__(self, cmd):
        """
        :param str cmd: A shell command to be executed
        """
        self.exception = None
        Executor.__init__(self, cmd)
        Thread.__init__(self)

    def join(self, timeout=None):
        """
        Ensure that both the thread and the subprocess have finished, and return self.proc's exit status.
        :param int timeout: As Thread.join
        """
        super().join(timeout=timeout)
        if self.exception:
            self._stop()
            self.error('Encountered a %s error: %s', self.exception.__class__.__name__, self.exception)
            raise EGCGError('Command failed: ' + self.cmd) from self.exception

        return self.proc.wait()

    def run(self):
        try:
            self._stream_output()
        except Exception as e:
            self.exception = e

    def _stream_output(self):
        """Run self._process and log its stdout/stderr until an EOF."""
        proc = self._process()
        read_set = [proc.stdout, proc.stderr]
        while read_set:
            rlist, wlist, xlist = select(read_set, [], [])

            for stream, emit in ((proc.stdout, self.info), (proc.stderr, self.error)):
                if stream in rlist:
                    line = stream.readline().decode('utf-8').strip()
                    if line:
                        emit(line)
                    else:
                        stream.close()
                        read_set.remove(stream)



import os
import pytest
import shutil
import subprocess
from unittest.mock import patch, Mock
from tests import TestEGCG
from egcg_core import executor
from egcg_core.executor.cluster_executor import ClusterExecutor, running_executors
from egcg_core.exceptions import EGCGError

get_stdout = 'egcg_core.executor.cluster_executor.ClusterExecutor._get_stdout'
sleep = 'egcg_core.executor.cluster_executor.sleep'


class TestExecutor(TestEGCG):
    executor_cls = executor.Executor

    def execute(self, *args, **kwargs):
        e = self.executor_cls(*args, **kwargs)
        e.start()
        return e.join()

    def test_cmd(self):
        assert self.execute('ls ' + os.path.dirname(self.assets_path)) == 0

    def test_dodgy_cmd(self):
        with patch.object(self.executor_cls, 'error'):
            assert self.execute('dodgy_cmd') == 127  # command not found

    def test_script(self):
        assert self.execute(os.path.join(self.assets_path, 'countdown.sh')) == 0

    def test_dodgy_script(self):
        # same exit status as the running script
        assert self.execute('%s dodgy' % os.path.join(self.assets_path, 'countdown.sh')) == 13

    def test_internal_error(self):
        with patch.object(self.executor_cls, 'info', side_effect=ValueError('Something went wrong')):
            with pytest.raises(EGCGError) as err:
                self.execute('ls')

            assert 'Command failed: ls' in str(err)

    def test_process(self):
        e = self.executor_cls('ls ' + os.path.dirname(self.assets_path))
        assert e.proc is None
        proc = e._process()
        assert proc is e.proc and isinstance(e.proc, subprocess.Popen)

    def test_bash_syntax(self):
        with patch.object(self.executor_cls, 'info') as mocked_info:
            assert self.execute('ls -lh %s | grep __init__' % os.path.dirname(self.assets_path)) == 0

        assert '__init__.py' in mocked_info.call_args_list[1][0][0]

    @patch('egcg_core.executor.local_execute', return_value=0)
    @patch('egcg_core.executor.cluster_execute', return_value=1)
    def test_main_method(self, mocked_cluster, mocked_local):
        assert executor.execute('this', 'that', env='local', some_other_arg='other') == 0
        mocked_local.assert_called_with('this', 'that')

        original_env = executor.cfg['executor']['job_execution']
        executor.cfg.content['executor']['job_execution'] = 'slurm'
        assert executor.execute('this', 'that', prelim_cmds=['other'], mem=4) == 1
        mocked_cluster.assert_called_with('this', 'that', env='slurm', prelim_cmds=['other'], mem=4)
        executor.cfg.content['executor']['job_execution'] = original_env


class TestStreamExecutor(TestExecutor):
    executor_cls = executor.StreamExecutor


class TestArrayExecutor(TestExecutor):
    executor_cls = executor.ArrayExecutor

    def execute(self, *args, **kwargs):
        return super().execute(args, stream=True, **kwargs)

    def test_cmd(self):
        e = self.executor_cls(['ls', 'ls -lh', 'pwd'], stream=True)
        e.start()
        assert e.join() == 0
        assert e.exit_statuses == [0, 0, 0]

    def test_dodgy_cmd(self):
        e = self.executor_cls(['ls', 'non_existent_cmd', 'pwd'], stream=True)
        for s in e.executors:
            s.error = Mock()

        e.error = Mock()
        e.start()
        assert e.join() == 127
        assert e.exit_statuses == [
            0,  # ls
            127,  # command not found
            0  # pwd
        ]

    def test_internal_error(self):
        with patch.object(executor.StreamExecutor, 'info', side_effect=ValueError('Something went wrong')):
            with pytest.raises(EGCGError) as err:
                self.execute('ls')

            assert 'Commands failed' in str(err)

    def test_process(self):
        e = self.executor_cls(['ls ' + os.path.dirname(self.assets_path)], stream=True)
        assert e.proc is None
        proc = e._process()
        assert proc is e.proc and isinstance(e.proc, subprocess.Popen)

    def test_bash_syntax(self):
        with patch.object(executor.StreamExecutor, 'info') as mocked_info:
            assert self.execute('ls -lh %s | grep __init__' % os.path.dirname(self.assets_path)) == 0

        assert '__init__.py' in mocked_info.call_args_list[1][0][0]


class TestClusterExecutor(TestEGCG):
    script = os.path.join(TestEGCG.assets_path, 'countdown.sh')
    executor_cls = ClusterExecutor

    def setUp(self):
        self.working_dir = os.path.join(self.assets_path, 'a_working_dir')
        os.makedirs(self.working_dir, exist_ok=True)
        self.executor = self.executor_cls(self.script, job_name='test_job', working_dir=self.working_dir)

    def tearDown(self):
        shutil.rmtree(self.working_dir)

    def test_get_stdout(self):
        popen = 'egcg_core.executor.executor.subprocess.Popen'
        with patch(popen, return_value=Mock(wait=Mock(return_value=None))) as p:
            assert self.executor._get_stdout('ls -d ' + self.assets_path).endswith('tests/assets')
            p.assert_called_with(['ls', '-d', self.assets_path], stdout=-1, stderr=-1)

    def test_run_and_retry(self):
        with patch(get_stdout, side_effect=[None, None, self.assets_path]) as p, patch(sleep):
            assert self.executor._run_and_retry('ls -d ' + self.assets_path).endswith('tests/assets')
            assert p.call_count == 3

    def test_dodgy_cmd(self):
        with pytest.raises(EGCGError) as err, patch(get_stdout, return_value=None), patch(sleep):
            self.executor.cmds = [os.path.join(self.assets_path, 'non_existent_script.sh')]
            self.executor.start()

        assert str(err).endswith('Job submission failed')

    def test_join(self):
        patched_job_finished = patch.object(self.executor_cls, '_job_finished', return_value=True)
        patched_exit_code = patch.object(self.executor_cls, '_job_exit_code', return_value=0)
        self.executor.finished_statuses = 'FXM'
        with patched_job_finished, patched_exit_code, patch(sleep):
            assert self.executor.join() == 0

    def test_job_cancellation(self):
        with patch.object(self.executor_cls, '_submit_job'),\
             patch.object(self.executor_cls, '_job_finished', return_value=True),\
             patch.object(self.executor_cls, 'write_script'),\
             patch.object(self.executor_cls, '_job_exit_code', return_value=9),\
             patch.object(self.executor_cls, 'cancel_job'), patch(sleep):

            self.executor.job_id = 'test_job'
            self.executor.start()
            assert running_executors == {'test_job': self.executor}
            executor.stop_running_jobs()
            assert running_executors == {}


class TestSlurmExecutor(TestClusterExecutor):
    executor_cls = executor.SlurmExecutor

    def test_sacct(self):
        with patch(get_stdout, return_value=' COMPLETED  0:0 \n COMPLETED  0:0\n FAILED 1:0') as p:
            assert self.executor._sacct('State,ExitCode') == {'COMPLETED  0:0', 'FAILED 1:0'}
            p.assert_called_with('sacct -nX -j None -o State,ExitCode')

    def test_squeue(self):
        with patch(get_stdout, return_value='RUNNING\nRUNNING\nRUNNING') as p:
            assert self.executor._squeue() == {'RUNNING'}
            p.assert_called_with('squeue -h -j None -o %T')

    def test_job_finished(self):
        sacct = 'egcg_core.executor.cluster_executor.SlurmExecutor._sacct'
        patched_squeue = patch('egcg_core.executor.cluster_executor.SlurmExecutor._squeue', return_value='')
        with patch(sacct, return_value=['RUNNING']), patched_squeue:
            assert not self.executor._job_finished()
        with patch(sacct, return_value=['COMPLETED']), patched_squeue:
            assert self.executor._job_finished()

    def test_job_exit_code(self):
        sacct = 'egcg_core.executor.cluster_executor.SlurmExecutor._sacct'
        with patch(sacct, return_value=['CANCELLED 0:0']):
            assert self.executor._job_exit_code() == 9
        with patch(sacct, return_value=['COMPLETED 0:x']):
            assert self.executor._job_exit_code() == 0



from os.path import join
from egcg_core.config import cfg
from egcg_core.app_logging import AppLogger
from egcg_core.exceptions import EGCGError


class ScriptWriter(AppLogger):
    """
    Writes a basic job submission script. Subclassed by SlurmWriter. Initialises with self.lines as an empty list, which
    is appended by self.write_line. This list is then saved line by line to self.script_file by self.save.
    """
    suffix = '.sh'
    array_index = 'JOB_INDEX'
    mapping = {
        'cpus': '# cpus: {}',
        'exclusive': '# exclusive',
        'job_name': '# job name: {}',
        'job_queue': '# queue: {}',
        'jobs': '# job array: 1-{}',
        'log_file': '# log file: {}',
        'mem': '# mem: {}gb',
        'walltime': '# walltime: {}',
    }

    def __init__(self, job_name, working_dir, job_queue=None, log_commands=True, **kwargs):
        self.log_commands = log_commands
        self.working_dir = working_dir
        self.log_file = join(self.working_dir, job_name + '.log')
        self.parameters = dict(
            kwargs,
            job_name=job_name,
            job_queue=job_queue or cfg['executor']['job_queue'],
            log_file=self.log_file
        )

        self.script_name = join(working_dir, job_name + self.suffix)
        self.debug('Writing script: %s', self.script_name)
        self.lines = []

    def register_cmd(self, cmd, log_file=None):
        if log_file:
            cmd += ' > %s 2>&1' % log_file
        self.add_line(cmd)

    def register_cmds(self, *cmds, parallel):
        if parallel:
            self.add_job_array(*cmds)
        else:
            self.lines.extend(list(cmds))

    def add_job_array(self, *cmds):
        if self.parameters.get('jobs'):
            raise EGCGError('Already written a job array - can only have one per script')

        if len(cmds) == 1:
            self.register_cmd(cmds[0])
        else:
            self._start_array()
            for idx, cmd in enumerate(cmds):
                self._register_array_cmd(
                    idx + 1,
                    cmd,
                    log_file=self.log_file + str(idx + 1) if self.log_commands else None
                )
            self._finish_array()
            self.parameters['jobs'] = len(cmds)

    def _register_array_cmd(self, idx, cmd, log_file=None):
        """
        :param int idx: The index of the job, i.e. which number the job has in the array
        :param str cmd: The command to write
        """
        line = str(idx) + ') ' + cmd
        if log_file:
            line += ' > ' + log_file + ' 2>&1'
        line += '\n' + ';;'
        self.add_line(line)

    def add_line(self, line):
        self.lines.append(line)

    def _start_array(self):
        self.add_line('case $%s in' % self.array_index)

    def _finish_array(self):
        self.add_line('*) echo "Unexpected %s: $%s"' % (self.array_index, self.array_index))
        self.add_line('esac')

    def line_break(self):
        self.lines.append('')

    def save(self):
        """Save self.lines to self.script_name."""
        with open(self.script_name, 'w') as f:
            f.write('\n'.join(self.lines) + '\n')

    def add_header(self):
        """Write a header for a given resource manager. If multiple jobs, split them into a job array."""
        header_lines = ['#!/bin/bash\n']
        for k in sorted(self.parameters):
            v = self.parameters[k]
            header_lines.append(self.mapping[k].format(v))

        header_lines.extend(['', 'cd ' + self.working_dir, ''])

        # prepend the formatted header
        self.lines = header_lines + self.lines


class SlurmWriter(ScriptWriter):
    """Writes a Bash script runnable on Slurm"""
    suffix = '.slurm'
    array_index = 'SLURM_ARRAY_TASK_ID'
    mapping = {
        'cpus': '#SBATCH --cpus-per-task={}',
        'exclusive': '#SBATCH --exclusive',
        'job_name': '#SBATCH --job-name="{}"',
        'job_queue': '#SBATCH --partition={}',
        'jobs': '#SBATCH --array=1-{}',
        'log_file': '#SBATCH --output={}',
        'mem': '#SBATCH --mem={}g',
        'walltime': '#SBATCH --time={}:00:00'
    }
